//---------------------------------------------------------------------------
// AnisotropicDisperse
//---------------------------------------------------------------------------
#if !defined(AnisotropicDisperse_H)
  #define AnisotropicDisperse_H

#include "GridBasedSpatialDisperse.h"

class DOMDocument;
class clSimManager;

/**
  Anisotropic spatial dispersal - Version 1.0
  <p>
  This behavior creates and disperses seeds according to the spatially-explicit
  model.  For each grid cell, the number of seeds is calculated based on the
  size and distance of parents of that species.  The seed distribution is
  anisotropic (meaning that seeds disperse farther in some directions than
  others).  The direction and magnitude of the maximum disperse distance is
  controlled on a per-species basis.  The shape of the seedling distribution
  distance in any given direction can be controlled by either a weibull or
  lognormal probability distribution function.
  <p>
  The "seeds" are actually tallies by species in a grid.  This behavior does
  not produce any seedlings.  A separate recruitment behavior must "germinate"
  the seeds into seedlings according to its own rules.
  <p>
  To call this in the parameter file, use "anisotropic disperse".  The
  namestring for this class is "anisotropic disperse".
  <p>
  Copyright 2003 Charles D. Canham.
  @author Lora E. Murphy

  <br>Edit history:
  <br>-----------------
  <br>April 28, 2004 - Submitted as beta (LEM)
*/
class clAnisotropicDisperse : virtual public clGridBasedSpatialDisperse {
//note: need the virtual keyword to avoid base class ambiguity.

 public:

/**
  Constructor.

  @param p_oSimManager Sim Manager object.
*/
 clAnisotropicDisperse(clSimManager *p_oSimManager);

/**
  Destructor.
*/
 ~clAnisotropicDisperse();

 protected:

 float ***mp_fProb;  /**<Normalized probability array for each species.
       Array size is # species by angle in 0.087 rad (5 degree) increments by
       max search radius.  These values are calculated in CalculateProbs().*/
 float *mp_fDirectionMaxDispersal; /**<Azimuth direction, in radians, of the
       max dispersal for each species.  Only used for anisotropic disperse.*/
 float *mp_fAnisotropicAmplitude; /**<Amplitude of the anisotropic effect for
       each species.  Only used for anisotropic disperse.*/
 float *mp_fNormalizer; /**<For normalizing values.*/
 float m_fAngleIncr; /**<We're rounding all angles to the nearest of this
       increment, in radians*/
 int m_iMaxSearchRadius; /**<Integer version of max search radius - rounded
       up to next integer - for sizing mp_fProb array*/
 int m_iNumAngles; /**<Number of angles in the mp_fProb array - for the
     destructor*/


/**
* Performs additional setup by getting values from the parameter file and
* calculating the probability array.
*/
  void DoChildSetup(xercesc::DOMDocument *p_oDoc);

  /**
  * Gets additional data from the parameter file.
  * @throw modelErr if the direction of max dispersal is not a valid angle in
  * radians.
  */
  void GetParameterData(xercesc::DOMDocument *p_oDoc);


/**
 * Gets the value in the mp_fProb array corresponding to the distance and angle
 * between the two points.  If the value is beyond the max search radius (and
 * thus beyond the array) 0 is returned.  The distance is RandomRound()ed to an
 * integer to get a value from the array.
 * @param fCellX X coordinate of point at center of grid cell into which
 * seeds are being dispersed.
 * @param fCellY Y coordinate of point at center of grid cell into which
 * seeds are being dispersed.
 * @param fTreeX X coordinate of neighbor parent tree.
 * @param fTreeY Y coordinate of neighbor parent tree.
 * @param iSpecies Species of neighbor parent tree.
 * @return Function value.
 */
 float GetFunctionValue(float &fCellX, float &fCellY, float &fTreeX,
 float &fTreeY, int &iSpecies);

/**
* Calculates the normalizer term.  The normalizer is the integration of the
* exponential portion of either the weibull or lognormal function.  The integral
* is calculated using a brute-force approach by summing the values of the
* function at successive values out to the maximum disperse distance, and at
* successive angles around a circle.
* @param iSpecies Species for which to calculate normalizer.
* @return Normalizer value.
*/
  float CalculateNormalizer(int iSpecies);

/**
* Calculates the normalized probability array.
*/
  void CalculateProbs();
};
//---------------------------------------------------------------------------
#endif // AnisotropicDisperse_H
