//---------------------------------------------------------------------------
// DensitySeedSurvival
//---------------------------------------------------------------------------
#if !defined(DensitySeedSurvival_H)
  #define DensitySeedSurvival_H

#include "BehaviorBase.h"

class clGridBase;
class clSimManager;

/**
* Density-Dependent Seed Survival - version 1.0
*
* This behavior reduces the number of seeds due to density-dependent effects.
*
* The number of seeds this behavior leaves of a given species in a particular
* "Dispersed Seeds" grid cell is:
* @htmlonly
  <br><center>R<sub>sp</sub> = S<sub>sp</sub>*exp(-c*S<sub>sp</sub><sup>&delta;</sup>)</center><br>
  @endhtmlonly
  where:</p>
* <ul>
* <li>R<sub>sp</sub> is the number of seedlings of a given species in the seed
* grid cell (competition is conspecific only)</li>
* <li>S<sub>sp</sub> is the potential number of seedlings of the species, i.e.
* the number of seeds dispersed to that grid cell
* <li>c is an input parameter - density-dependent slope</li>
* <li>@htmlonly &delta; @endhtmlonly is an input parameter - density-dependent
* steepness</li>
* </ul>
*
* A fatal error is thrown if a disperse behavior is not also defined for the
* run.
*
* This class's namestring and parameter call string are both
* "Density Dependent Seed Survival".
* Copyright 2005 Charles D. Canham.
* @author Lora E. Murphy
*
* <br>Edit history:
* <br>-----------------
* <br>April 12, 2005 - Created (LEM)
*/
class clDensitySeedSurvival : virtual public clBehaviorBase {
//note: need the virtual keyword to avoid base class ambiguity.

 public:

  /**
  * Constructor.
  *
  * @param p_oSimManager Sim Manager object.
  */
  clDensitySeedSurvival(clSimManager *p_oSimManager);

  /**
  * Destructor.
  */
  ~clDensitySeedSurvival();

  /**
  * Reads in values from the parameter file.
  *
  * @param p_oDoc DOM tree of parsed input file.
  */
  void GetData(xercesc::DOMDocument *p_oDoc);

  /**
  * Performs density-dependent seed survival.  For each grid cell in the seed
  * grid, for each species to which this behavior applies, the number of seeds
  * left is calculated by assessing the equation above. A random round is used
  * to take care of fractional parts of seeds.
  */
  void Action();

  protected:

  /**Pointer to the "Dispersed Seeds" grid created by disperse behaviors.*/
  clGridBase *mp_oSeedGrid;

  /**Density-dependence steepness parameter.  Array size is total number of
  * species.*/
  float *mp_fDensDepSteepness;

  /**Density-dependence slope parameter.  Array size is total number of
  * species.*/
  float *mp_fDensDepSlope;

  /**Data member codes for seed grid for number of seeds.  Array size is #
  * total species.*/
  short int *mp_iSeedGridCode;
};
//---------------------------------------------------------------------------
#endif // Germination_H

