//---------------------------------------------------------------------------
// IsotropicDisperse
//---------------------------------------------------------------------------
#if !defined(IsotropicDisperse_H)
  #define IsotropicDisperse_H

#include "GridBasedSpatialDisperse.h"

class DOMDocument;
/**
  Isotropic spatial dispersal - Version 1.0
  <p>
  This behavior creates and disperses seeds according to the spatially-explicit
  model.  For each grid cell, the number of seeds is calculated based on the
  size and distance of parents of that species.  The seed distribution is
  isotropic (meaning that seeds disperse evenly in all directions).  The
  distribution of seedlings is controlled by either a weibull or
  lognormal probability distribution function.
  <p>
  The "seeds" are actually tallies by species in a grid.  This behavior does
  not produce any seedlings.  A separate recruitment behavior must "germinate"
  the seeds into seedlings according to its own rules.
  <p>
  To call this in the parameter file, use "isotropic disperse".  The
  namestring for this class is "isotropic disperse".
  <p>
  Copyright 2003 Charles D. Canham.
  @author Lora E. Murphy

  <br>Edit history:
  <br>-----------------
  <br>April 28, 2004 - Submitted as beta (LEM)
*/
class clIsotropicDisperse : virtual public clGridBasedSpatialDisperse {
//note: need the virtual keyword to avoid base class ambiguity.

 public:

/**
  Constructor.

  @param p_oSimManager Sim Manager object.
*/
 clIsotropicDisperse(clSimManager *p_oSimManager);

/**
  Destructor.
*/
 ~clIsotropicDisperse();

 protected:

 float **mp_fProb;  /**<Normalized probability array for each species.
       Array size is # species by max search radius.  These values are
       calculated in CalculateProbs().*/
 int m_iMaxSearchRadius; /**<Integer version of max search radius - rounded
       up to next integer - for sizing mp_fProb array*/

 /**
  * Gets the value in the mp_fProb array corresponding to the distance between
  * the two points.  If the value is beyond the max search radius (and thus
  * beyond the array) 0 is returned.  The distance is RandomRound()ed to an
  * integer to get a value from the array.
  * @param fCellX X coordinate of point at center of grid cell into which
  * seeds are being dispersed.
  * @param fCellY Y coordinate of point at center of grid cell into which
  * seeds are being dispersed.
  * @param fTreeX X coordinate of neighbor parent tree.
  * @param fTreeY Y coordinate of neighbor parent tree.
  * @param iSpecies Species of neighbor parent tree.
  * @return Function value.
  */
  float GetFunctionValue(float &fCellX, float &fCellY, float &fTreeX,
  float &fTreeY, int &iSpecies);

/**
* Performs additional setup by declaring the mp_fProb array and populating its
* values with CalculateProbs().
*/
  void DoChildSetup(xercesc::DOMDocument *p_oDoc);

  /**
  * Calculates the normalized probability array for each species.  This is not
  * a cumulative probability array - just the value of the exponential of
  * either the weibull or lognormal function divided by the normalizer.  This
  * normalizer is the sum of all values of the function out to the maximum
  * dimension of the grid, with a max of 1000.
  */
  void CalculateProbs();

  /**
  * Calculates the normalizer term.  It is calculated in increments of 0.25 m2
  * and then divided by 4 to restore it to units of m2.
  * @param iSpecies Species for which to calculate normalizer.
  */
  float CalculateNormalizer(int iSpecies);

};
//---------------------------------------------------------------------------
#endif // AnisotropicDisperse_H
