//---------------------------------------------------------------------------
#include "Output.h"
#include "GridBase.h"
#include "SimManager.h"
#include "Plot.h"
#include "TreePopulation.h"
#include "ParsingFunctions.h"
#include "PlatformFuncs.h"

const int BUFFER_SIZE = 1023;

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
//     Destructor
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

//////////////////////////////////////////////////////////////////////////////
clOutput::~clOutput()
{
  int i; //loop counter

  //Delete arrays
  for ( i = 0; i < m_iNumTreeSettingsToSave; i++ )
  {
    delete[] mp_treeSettings[i].p_iIntCodes;
    delete[] mp_treeSettings[i].p_iFloatCodes;
    delete[] mp_treeSettings[i].p_iCharCodes;
    delete[] mp_treeSettings[i].p_iBoolCodes;
  }
  delete[] mp_treeSettings; mp_treeSettings = NULL;
  for ( i = 0; i < m_iNumGridsToSave; i++ )
  {
    delete[] mp_gridSettings[i].p_iIntCodes;
    delete[] mp_gridSettings[i].p_iFloatCodes;
    delete[] mp_gridSettings[i].p_iCharCodes;
    delete[] mp_gridSettings[i].p_iBoolCodes;
    delete[] mp_gridSettings[i].p_iPackageIntCodes;
    delete[] mp_gridSettings[i].p_iPackageFloatCodes;
    delete[] mp_gridSettings[i].p_iPackageCharCodes;
    delete[] mp_gridSettings[i].p_iPackageBoolCodes;
  }
  delete[] mp_gridSettings;
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
//     Constructor
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

/////////////////////////////////////////////////////////////////////////////
clOutput::clOutput( clSimManager * p_oSimManager ) : clBehaviorBase( p_oSimManager ), clWorkerBase( p_oSimManager )
{
  try
  {
    strcpy( m_cNameString, "output" );

    //Versions
    m_fVersionNumber = 1;
    m_fMinimumVersionNumber = 1;

    //Null pointers, initialize variables to empty
    mp_treeSettings = NULL;
    mp_gridSettings = NULL;
    m_iNumTreeSettingsToSave = 0;
    m_iNumGridsToSave = 0;
    strcpy( m_cFileRoot, "" );

    //Allowed file types
    m_iNumAllowedTypes = 2;
    mp_iAllowedFileTypes = new int[m_iNumAllowedTypes];
    mp_iAllowedFileTypes[0] = parfile;
    mp_iAllowedFileTypes[1] = detailed_output;
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::clOutput" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
//     GetData()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

/////////////////////////////////////////////////////////////////////////////
void clOutput::GetData( DOMDocument * p_oDoc )
{
  try
  {
    DOMElement * p_oElement; //for casting to DOM_Element
    char cTemp[MAX_FILENAME_SIZE];
    int i, j, k, iNumTimesteps, iMaxRootFileSize, iStrLen;
    bool bTest;

    //General data and flags

    //Filenames
    //get the doc root element
    p_oElement = p_oDoc->getDocumentElement();
    FillSingleValue( p_oElement, "ou_filename", cTemp, MAX_FILENAME_SIZE, true );

    strcpy( m_cFileRoot, cTemp );

    //Strip the filename of the tarball extension, if it's there, to get the
    //file root
    //Start by making cTemp all lower case
    for ( i = 0; i < strlen( cTemp ); i++ )
    {
      cTemp[i] = tolower( cTemp[i] );
    }

    //Work backwards the length of the file extension in cTemp and compare
    if ( strlen( cTemp ) > strlen( TARBALL_FILE_EXT ) )
    {

      bTest = true;
      j = 0;
      for ( i = strlen( cTemp ) - strlen( TARBALL_FILE_EXT ); i < strlen( cTemp ); i++ )
      {
        if ( cTemp[i] != TARBALL_FILE_EXT[j] )
        {
          bTest = false;
          break;
        }
        j++;
      }

      if ( bTest == true )
      {

        //The file extension's already on there - so strip the extension
        //off of root
        strcpy( m_cTarball, m_cFileRoot );
        m_cFileRoot[strlen( m_cFileRoot ) - strlen( TARBALL_FILE_EXT )] = '\0';

      }
    }

    //We want all our filenames to fit into MAX_FILENAME_SIZE.  So we have to
    //make our root short enough to accommodate the timestep tack-ons.  Get the
    //number of timesteps and calculate the max length of our root filename,
    //which is max filename length minus number of digis in max timesteps minus
    //1 for the underscore minus the length of the zipped XML extension
    //(The -2 is one for the underscore and one because log10 gives us 1 less
    //than the number of digits in timesteps)
    iNumTimesteps = mp_oSimManager->GetNumberOfTimesteps();
    iMaxRootFileSize = MAX_FILENAME_SIZE - floor( log10( iNumTimesteps ) ) - 2 - strlen( ZIPPED_FILE_EXT ) - 1; //one for the NULL

    //If our file length is over that value, we have to truncate it.  If we have
    //to truncate down through the folder level, we'd better throw an error.
    if ( strlen( m_cFileRoot ) > iMaxRootFileSize )
    {

      //Look for a slash - forward or backward - this should cover most OS
      //file structures
      for ( i = iMaxRootFileSize; i < strlen( m_cFileRoot ); i++ )
      {
        if ( '\\' == m_cFileRoot[i] || '/' == m_cFileRoot[i] )
        {
          modelErr stcErr;
          stcErr.iErrorCode = BAD_FILE;
          strcpy( stcErr.cFunction, "clOutput::GetData" );
          strcpy( stcErr.cMoreInfo, "Detailed output file name is too long." );
          throw( stcErr );
        }
      }

      //If we made it this far, we don't have to shave off any directories.
      //So just truncate the root.
      m_cFileRoot[iMaxRootFileSize] = '\0';
    }

    //Make the tarball file name by appending the tarball extension to the root
    sprintf( m_cTarball, "%s%s", m_cFileRoot, TARBALL_FILE_EXT );

    //Now take off the directory structure so that TAR won't preserve a
    //directory structure for files - which is good since very long filenames
    //confound TAR otherwise
    strcpy(cTemp, m_cFileRoot);
    iStrLen = strlen(m_cFileRoot) - 1;
    for (i = iStrLen; i >= 0; i--) {
      if ( '\\' == cTemp[i] || '/' == cTemp[i] ) {
        k = 0;
        iStrLen++;
        for (j = i+1; j < iStrLen; j++) {
          m_cFileRoot[k] = cTemp[j];
          k++;
        }
        m_cFileRoot[k] = '\0';
        break;
      }
    }

    ExtractTreeInfo( p_oDoc );
    ExtractGridInfo( p_oDoc );

    //Just double-check that we have settings to save - if not, erase the
    //filename
    if ( 0 == m_iNumTreeSettingsToSave && 0 == m_iNumGridsToSave )
      strcpy( m_cFileRoot, "" );

    //Make sure that everything's in place for zipping - if not throw error
    if ( !TarballSetup(mp_oSimManager->GetAppPath()) )
    {
      modelErr stcErr;
      stcErr.iErrorCode = CANT_FIND_OBJECT;
      strcpy( stcErr.cMoreInfo, "Missing file gzip.exe and/or tar.exe" );
      strcpy( stcErr.cFunction, "clOutput::GetData" );
      throw( stcErr );
    }

    //Write the output file headers
    WriteDetailedOutputHeader();

    //Write the initial conditions timestep
    Action();

  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::GetData" );
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
//     ExtractTreeInfo()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

/////////////////////////////////////////////////////////////////////////////
void clOutput::ExtractTreeInfo( DOMDocument * p_oDoc )
{
  try
  {
    clPopulationBase * p_oTemp;
    clTreePopulation * p_oPop;
    DOMNodeList * p_oSettingsList, //holds the list of tree info structures
         * p_oDataMemberList; //holds list of tree data members
    DOMNode * p_oDocNode;
    DOMElement * p_oTreeInfo, //one tree type's info
         * p_oDataMember; //one data member
    char * cData; //for extracting data from the file
    short int iNumSettings, //how many types we have output data for in the file
         iSpecies, iType, //species number
         iDataMemberCode, //data member code for a label
         i, j; //loop counters

    //Get the tree population
    p_oTemp = mp_oSimManager->GetPopulationObject( "treepopulation" );
    p_oPop = ( clTreePopulation * ) p_oTemp;

    /** Look for tree info */
    //Search for all tree type info sets
    p_oSettingsList = p_oDoc->getElementsByTagName( XMLString::transcode( "ou_treeInfo" ) );
    iNumSettings = p_oSettingsList->getLength();

    //If we found nothing, exit with no error
    if ( iNumSettings == 0 ) return;

    //Allocate the grid memory and initialize - set frequencies to -1 so we
    //can distinguish absence of data from the user setting 0
    m_iNumTreeSettingsToSave = iNumSettings;
    mp_treeSettings = new stcTreeOutputInfo[m_iNumTreeSettingsToSave];
    for ( i = 0; i < m_iNumTreeSettingsToSave; i++ )
    {
      mp_treeSettings[i].iType = -1;
      mp_treeSettings[i].iSpecies = -1;
      mp_treeSettings[i].iSaveFreq = -1;
      mp_treeSettings[i].iSumTimestep = 0;
      mp_treeSettings[i].iNumInts = 0;
      mp_treeSettings[i].iNumFloats = 0;
      mp_treeSettings[i].iNumChars = 0;
      mp_treeSettings[i].iNumBools = 0;
      mp_treeSettings[i].p_iIntCodes = NULL;
      mp_treeSettings[i].p_iFloatCodes = NULL;
      mp_treeSettings[i].p_iCharCodes = NULL;
      mp_treeSettings[i].p_iBoolCodes = NULL;
    }

    //go through each type and break out its info
    for ( i = 0; i < m_iNumTreeSettingsToSave; i++ )
    {
      p_oDocNode = p_oSettingsList->item( i );
      p_oTreeInfo = ( DOMElement * ) p_oDocNode;

      //get the type name - if something goes wrong throw an error
      cData = XMLString::transcode( p_oTreeInfo->getAttributeNode( XMLString::transcode( "type" ) )->getNodeValue() );
      if ( strcmp( "Seed", cData ) == 0 )
        iType = clTreePopulation::seed;
      else if ( strcmp( "Seedling", cData ) == 0 )
        iType = clTreePopulation::seedling;
      else if ( strcmp( "Sapling", cData ) == 0 )
        iType = clTreePopulation::sapling;
      else if ( strcmp( "Adult", cData ) == 0 )
        iType = clTreePopulation::adult;
      else if ( strcmp( "Snag", cData ) == 0 )
        iType = clTreePopulation::snag;
      else if ( strcmp( "Woody_debris", cData ) == 0 )
        iType = clTreePopulation::woody_debris;
      else
      {
        modelErr stcErr;
        strcpy( stcErr.cFunction, "clOutput::ExtractTreeInfo" );
        stcErr.iErrorCode = BAD_DATA;
        sprintf( stcErr.cMoreInfo, "%s%s", "unrecognized type - ", cData );
        throw( stcErr );
      }

      //Species
      cData = XMLString::transcode( p_oTreeInfo->getAttributeNode( XMLString::transcode( "species" ) )->getNodeValue() );
      iSpecies = p_oPop->TranslateSpeciesNameToCode( cData );
      if ( -1 == iSpecies )
      { //unrecognized species - throw error
        modelErr stcErr;
        stcErr.iErrorCode = BAD_DATA;
        strcpy( stcErr.cFunction, "clOutput::ExtractTreeInfo" );
        sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized species - ", cData );
        throw( stcErr );
      }

      //Check to see if this species/type combo is a repeat - if so, throw error
      for ( j = 0; j < i; j++ )
      {
        if ( mp_treeSettings[i].iType == mp_treeSettings[j].iType
             && mp_treeSettings[i].iSpecies == mp_treeSettings[j].iSpecies )
             {
               modelErr stcErr;
               stcErr.iErrorCode = BAD_DATA;
               strcpy( stcErr.cFunction, "clOutput::ExtractTreeInfo" );
               strcpy( stcErr.cMoreInfo, "Repeat species/type combo." );
               throw( stcErr );
        }
      }

      mp_treeSettings[i].iType = iType;
      mp_treeSettings[i].iSpecies = iSpecies;

      //Frequency data
      FillSingleValue( p_oTreeInfo, "ou_saveFreq", ( int * ) & mp_treeSettings[i].iSaveFreq, true );

      //Make sure that the frequency is not negative - if so, throw error
      if ( mp_treeSettings[i].iSaveFreq < 0 )
      {
        modelErr stcErr;
        stcErr.iErrorCode = BAD_DATA;
        strcpy( stcErr.cFunction, "clOutput::ExtractTreeInfo" );
        strcpy( stcErr.cMoreInfo, "Save frequency is negative" );
        throw( stcErr );
      }

      //Number to save for each data member type
      //****************Ints*******************
      p_oDataMemberList = p_oTreeInfo->getElementsByTagName( XMLString::transcode( "ou_int" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_treeSettings[i].iNumInts = p_oDataMemberList->getLength();
        //Declare the int array
        mp_treeSettings[i].p_iIntCodes = new short int[mp_treeSettings[i].iNumInts];
        for ( j = 0; j < mp_treeSettings[i].iNumInts; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = p_oPop->GetIntDataCode( cData, iSpecies, iType );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractTreeInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized int data member label - ", cData );
            throw( stcErr );
          }
          mp_treeSettings[i].p_iIntCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())
      //****************Floats*******************
      p_oDataMemberList = p_oTreeInfo->getElementsByTagName( XMLString::transcode( "ou_float" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_treeSettings[i].iNumFloats = p_oDataMemberList->getLength();
        //Declare the float array
        mp_treeSettings[i].p_iFloatCodes = new short int[mp_treeSettings[i].iNumFloats];
        for ( j = 0; j < mp_treeSettings[i].iNumFloats; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = p_oPop->GetFloatDataCode( cData, iSpecies, iType );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractTreeInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized float data member label - ", cData );
            throw( stcErr );
          }
          mp_treeSettings[i].p_iFloatCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())
      //****************Chars*******************
      p_oDataMemberList = p_oTreeInfo->getElementsByTagName( XMLString::transcode( "ou_char" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_treeSettings[i].iNumChars = p_oDataMemberList->getLength();
        //Declare the int array
        mp_treeSettings[i].p_iCharCodes = new short int[mp_treeSettings[i].iNumChars];
        for ( j = 0; j < mp_treeSettings[i].iNumChars; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = p_oPop->GetCharDataCode( cData, iSpecies, iType );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractTreeInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized char data member label - ", cData );
            throw( stcErr );
          }
          mp_treeSettings[i].p_iCharCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())
      //****************Bools*******************
      p_oDataMemberList = p_oTreeInfo->getElementsByTagName( XMLString::transcode( "ou_bool" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_treeSettings[i].iNumBools = p_oDataMemberList->getLength();
        //Declare the int array
        mp_treeSettings[i].p_iBoolCodes = new short int[mp_treeSettings[i].iNumBools];
        for ( j = 0; j < mp_treeSettings[i].iNumBools; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = p_oPop->GetBoolDataCode( cData, iSpecies, iType );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractTreeInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized bool data member label - ", cData );
            throw( stcErr );
          }
          mp_treeSettings[i].p_iBoolCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())
    } //end of for (i = 0; i < m_iNumTypeToSave; i++)
  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::ExtractTreeInfo" );
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
//     ExtractGridInfo()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

/////////////////////////////////////////////////////////////////////////////
void clOutput::ExtractGridInfo( DOMDocument * p_oDoc )
{
  try
  {
    DOMNodeList * p_oGridList, //list of grids
         * p_oDataMemberList; //list of data members
    DOMNode * p_oDocNode; //for retrieving data
    DOMElement * p_oGrid, //one grid element
         * p_oDataMember; //data member element
    char * cData; //for extracting text data
    short int iNumGrids, //number of grids found
         iNumPoints, //number of points found
         iDataMemberCode, //data member code
         i, j; //loop counters

    //Get all the grid info structures out of the file, if any
    p_oGridList = p_oDoc->getElementsByTagName( XMLString::transcode( "ou_gridInfo" ) );
    iNumGrids = p_oGridList->getLength();
    if ( 0 == iNumGrids ) return; //if nothing found exit without error

    //Allocate memory - set frequencies to -1 so we
    //can distinguish absence of data from the user setting 0
    m_iNumGridsToSave = iNumGrids;
    mp_gridSettings = new stcGridOutputInfo[m_iNumGridsToSave];
    for ( i = 0; i < m_iNumGridsToSave; i++ )
    {
      mp_gridSettings[i].iSaveFreq = -1;
      mp_gridSettings[i].iSumTimestep = 0;
      mp_gridSettings[i].iNumInts = 0;
      mp_gridSettings[i].iNumFloats = 0;
      mp_gridSettings[i].iNumChars = 0;
      mp_gridSettings[i].iNumBools = 0;
      mp_gridSettings[i].p_iIntCodes = NULL;
      mp_gridSettings[i].p_iFloatCodes = NULL;
      mp_gridSettings[i].p_iCharCodes = NULL;
      mp_gridSettings[i].p_iBoolCodes = NULL;
      mp_gridSettings[i].iNumPackageInts = 0;
      mp_gridSettings[i].iNumPackageFloats = 0;
      mp_gridSettings[i].iNumPackageChars = 0;
      mp_gridSettings[i].iNumPackageBools = 0;
      mp_gridSettings[i].p_iPackageIntCodes = NULL;
      mp_gridSettings[i].p_iPackageFloatCodes = NULL;
      mp_gridSettings[i].p_iPackageCharCodes = NULL;
      mp_gridSettings[i].p_iPackageBoolCodes = NULL;
    } //end of for (i = 0; i < m_iNumGridsToSave; i++)

    //Go through the grid objects in the file one at a time and extract the info
    for ( i = 0; i < m_iNumGridsToSave; i++ )
    {
      p_oDocNode = p_oGridList->item( i );
      p_oGrid = ( DOMElement * ) p_oDocNode;

      //First make sure that this is for a valid grid object
      cData = XMLString::transcode( p_oGrid->getAttributeNode( XMLString::transcode( "gridName" ) )->getNodeValue() );
      mp_gridSettings[i].p_oGridPointer = mp_oSimManager->GetGridObject( cData );
      if ( NULL == mp_gridSettings[i].p_oGridPointer )
      {
        modelErr stcErr;
        strcpy( stcErr.cFunction, "clOutput::ExtractGridInfo" );
        stcErr.iErrorCode = CANT_FIND_OBJECT;
        strcpy( stcErr.cMoreInfo, cData );
        throw( stcErr );
      }

      //Get frequency
      FillSingleValue( p_oGrid, "ou_saveFreq", ( int * ) & mp_gridSettings[i].iSaveFreq, true );

      //Number to save for each data member type
      //****************Ints*******************
      p_oDataMemberList = p_oGrid->getElementsByTagName( XMLString::transcode( "ou_int" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_gridSettings[i].iNumInts = p_oDataMemberList->getLength();
        //Declare the int array
        mp_gridSettings[i].p_iIntCodes = new short int[mp_gridSettings[i].iNumInts];
        for ( j = 0; j < mp_gridSettings[i].iNumInts; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = mp_gridSettings[i].p_oGridPointer->GetIntDataCode( cData );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractGridInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized int data member label - ", cData );
            throw( stcErr );
          }
          mp_gridSettings[i].p_iIntCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())
      //****************Floats*******************
      p_oDataMemberList = p_oGrid->getElementsByTagName( XMLString::transcode( "ou_float" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_gridSettings[i].iNumFloats = p_oDataMemberList->getLength();
        //Declare the float array
        mp_gridSettings[i].p_iFloatCodes = new short int[mp_gridSettings[i].iNumFloats];
        for ( j = 0; j < mp_gridSettings[i].iNumFloats; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = mp_gridSettings[i].p_oGridPointer->GetFloatDataCode( cData );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractGridInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized float data member label - ", cData );
            throw( stcErr );
          }
          mp_gridSettings[i].p_iFloatCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())
      //****************Chars*******************
      p_oDataMemberList = p_oGrid->getElementsByTagName( XMLString::transcode( "ou_char" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_gridSettings[i].iNumChars = p_oDataMemberList->getLength();
        //Declare the char array
        mp_gridSettings[i].p_iCharCodes = new short int[mp_gridSettings[i].iNumChars];
        for ( j = 0; j < mp_gridSettings[i].iNumChars; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = mp_gridSettings[i].p_oGridPointer->GetCharDataCode( cData );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractGridInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized char data member label - ", cData );
            throw( stcErr );
          }
          mp_gridSettings[i].p_iCharCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())
      //****************Bools*******************
      p_oDataMemberList = p_oGrid->getElementsByTagName( XMLString::transcode( "ou_bool" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_gridSettings[i].iNumBools = p_oDataMemberList->getLength();
        //Declare the bool array
        mp_gridSettings[i].p_iBoolCodes = new short int[mp_gridSettings[i].iNumBools];
        for ( j = 0; j < mp_gridSettings[i].iNumBools; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = mp_gridSettings[i].p_oGridPointer->GetBoolDataCode( cData );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractGridInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized bool data member label - ", cData );
            throw( stcErr );
          }
          mp_gridSettings[i].p_iBoolCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())


      //Packages
      //****************Package Ints*******************
      p_oDataMemberList = p_oGrid->getElementsByTagName( XMLString::transcode( "ou_packageInt" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_gridSettings[i].iNumPackageInts = p_oDataMemberList->getLength();
        //Declare the int array
        mp_gridSettings[i].p_iPackageIntCodes = new short int[mp_gridSettings[i].iNumPackageInts];
        for ( j = 0; j < mp_gridSettings[i].iNumPackageInts; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = mp_gridSettings[i].p_oGridPointer->GetPackageIntDataCode( cData );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractGridInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized package int data member label - ", cData );
            throw( stcErr );
          }
          mp_gridSettings[i].p_iPackageIntCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())
      //****************Package Floats*******************
      p_oDataMemberList = p_oGrid->getElementsByTagName( XMLString::transcode( "ou_packageFloat" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_gridSettings[i].iNumPackageFloats = p_oDataMemberList->getLength();
        //Declare the float array
        mp_gridSettings[i].p_iPackageFloatCodes = new short int[mp_gridSettings[i].iNumPackageFloats];
        for ( j = 0; j < mp_gridSettings[i].iNumPackageFloats; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = mp_gridSettings[i].p_oGridPointer->GetPackageFloatDataCode( cData );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractGridInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized package float data member label - ", cData );
            throw( stcErr );
          }
          mp_gridSettings[i].p_iPackageFloatCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())
      //****************Package Chars*******************
      p_oDataMemberList = p_oGrid->getElementsByTagName( XMLString::transcode( "ou_packageChar" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_gridSettings[i].iNumPackageChars = p_oDataMemberList->getLength();
        //Declare the char array
        mp_gridSettings[i].p_iPackageCharCodes = new short int[mp_gridSettings[i].iNumPackageChars];
        for ( j = 0; j < mp_gridSettings[i].iNumPackageChars; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = mp_gridSettings[i].p_oGridPointer->GetPackageCharDataCode( cData );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractGridInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized package char data member label - ", cData );
            throw( stcErr );
          }
          mp_gridSettings[i].p_iPackageCharCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())
      //****************Package Bools*******************
      p_oDataMemberList = p_oGrid->getElementsByTagName( XMLString::transcode( "ou_packageBool" ) );
      if ( 0 != p_oDataMemberList->getLength() )
      {
        mp_gridSettings[i].iNumPackageBools = p_oDataMemberList->getLength();
        //Declare the bool array
        mp_gridSettings[i].p_iPackageBoolCodes = new short int[mp_gridSettings[i].iNumPackageBools];
        for ( j = 0; j < mp_gridSettings[i].iNumPackageBools; j++ )
        {
          p_oDocNode = p_oDataMemberList->item( j );
          p_oDataMember = ( DOMElement * ) p_oDocNode;
          cData = XMLString::transcode( p_oDataMember->getFirstChild()->getNodeValue() );
          iDataMemberCode = mp_gridSettings[i].p_oGridPointer->GetPackageBoolDataCode( cData );
          if ( -1 == iDataMemberCode )
          { //unrecognized data code - throw error
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clOutput::ExtractGridInfo" );
            sprintf( stcErr.cMoreInfo, "%s%s", "Unrecognized package bool data member label - ", cData );
            throw( stcErr );
          }
          mp_gridSettings[i].p_iPackageBoolCodes[j] = iDataMemberCode;
        }
      } //end of if (0 != p_oDataMemberList->getLength())
    } //end of (i = 0; i < m_iNumGridsToSave; i++)
  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::ExtractGridInfo" );
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// WriteDetailedOutputHeader()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
void clOutput::WriteDetailedOutputHeader()
{
  try
  {
    const int BUFFER_SIZE = 128;

    FILE * outFile, //detailed output file
         * inFile; //parameter file
    char cFileCode[9], //holds the file code
         cFilename[MAX_FILENAME_SIZE], cBuf[BUFFER_SIZE]; //for transferring between files
    char * cPos;
    int iNumRead, //for testing for end-of-file
         iPos, i;
    bool bReading; //whether or not we found the <paramFile tag

    if ( strlen( m_cFileRoot ) == 0 ) return;

    //Tack on the file extension to the root
    sprintf( cFilename, "%s%s", m_cFileRoot, DETAILED_OUTPUT_FILE_EXT );

    //There is a possibility that the detailed output filename and the
    //parameter filename are the same - especially if a detailed output file
    //was loaded as a parameter file.  In this case, just skip out
    if ( strcmp( cFilename, mp_oSimManager->GetParFilename() ) == 0 )
      return;

    //Open detailed output file
    outFile = fopen( cFilename, "wt" ); //overwrite existing

    //If we can't open the file throw an error
    if ( NULL == outFile )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_FILE;
      strcpy( stcErr.cFunction, "clOutput::WriteDetailedOutputHeader" );
      sprintf( stcErr.cMoreInfo, "%s%s%s", "Couldn't open file \"", cFilename, "\"." );
      throw( stcErr );
    }

    //Open parameter file
    inFile = fopen( mp_oSimManager->GetParFilename(), "rt" );

    //If we can't open the file throw an error
    if ( NULL == inFile )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_FILE;
      strcpy( stcErr.cFunction, "clOutput::WriteDetailedOutputHeader" );
      sprintf( stcErr.cMoreInfo, "%s%s%s", "Couldn't open file \"", mp_oSimManager->GetParFilename(), "\"." );
      throw( stcErr );
    }

    //Get the file code string
    AssembleFileCode( detailed_output, DETAILED_OUTPUT_FILE_VERSION, cFileCode );

    //Write the detailed output opening tag
    sprintf( cBuf, "%s%s%s", "<?xml version=\"1.0\" encoding=\"UTF-8\"?><rundata fileCode=\"", cFileCode, "\">" );
    fwrite( cBuf, strlen( cBuf ), 1, outFile );

    //Now start the data feed from parameter file to detailed output file.  We
    //want to start at the <paramFile> tag.
    bReading = false;
    iNumRead = BUFFER_SIZE - 1;

    while ( iNumRead == BUFFER_SIZE - 1 )
    { //this is our end-of-file test
      iNumRead = fread( cBuf, sizeof( char ), BUFFER_SIZE - 1, inFile );
      cBuf[iNumRead] = '\0';

      if ( !bReading )
      {

        //We haven't found the parameter file opener tag yet - so find it first
        cPos = strstr( cBuf, "<paramFile" );
        if ( cPos != NULL )
        {

          //It's in this string - find the index in the string where cPos is
          iPos = -1;
          for ( i = 0; i < strlen( cBuf ); i++ )
          {
            if ( cPos == cBuf + i )
            {
              iPos = i;
              break;
            }
          }

          if ( iPos > -1 )
          {
            fwrite( cPos, strlen( cBuf ) - iPos, 1, outFile );
            bReading = true;
          }
        }

      }
      else
      {

        //We've already found the parameter string - so it's a straight
        //read/write
        fwrite( cBuf, strlen( cBuf ), 1, outFile );

      }
    }


    //Write detailed output closing tag

    strcpy( cBuf, "</rundata>" );
    fwrite( cBuf, strlen( cBuf ), 1, outFile );

    fclose( inFile );
    fclose( outFile );

    //Zip and tar the file
    ZipFile( cFilename, mp_oSimManager->GetAppPath() );
    strcat( cFilename, GZIP_EXT );
    AddFileToNewTarball( m_cTarball, cFilename, mp_oSimManager->GetAppPath() );
    DeleteThisFile( cFilename ); //it's copied to tarball - so remove original

  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::WriteDetailedOutputHeader" );
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
//     Action()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

/////////////////////////////////////////////////////////////////////////////
void clOutput::Action()
{
  try
  {
    if ( strlen( m_cFileRoot ) == 0 ) return;

    //Get the filename of the detailed output timestep file
    int iTimestep = mp_oSimManager->GetCurrentTimestep();
    char cTimestepFilename[MAX_FILENAME_SIZE];
    GetTimestepFilename( iTimestep, cTimestepFilename );

    //Now write the file
    WriteTimestepHeader( cTimestepFilename );
    WriteTreeData( cTimestepFilename );
    WriteGridData( cTimestepFilename );
    WriteTimestepFooter( cTimestepFilename );

    //Zip the file and add it to the tarball
    ZipFile( cTimestepFilename, mp_oSimManager->GetAppPath() );
    strcat( cTimestepFilename, GZIP_EXT ); //'cause that's the new filename after
    //zip
    AddFileToTarball( m_cTarball, cTimestepFilename, mp_oSimManager->GetAppPath() );
    DeleteThisFile( cTimestepFilename );

  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  } //throw without clearing this object
  catch ( modelMsg & msg )
  {
    throw( msg );
  }
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::Action" );
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
//     WriteTreeData()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

/////////////////////////////////////////////////////////////////////////////
void clOutput::WriteTreeData( const char * cFilename )
{
  try
  {
    clPopulationBase * p_oTemp; //for getting the tree population
    clTreePopulation * p_oTrees; //pointer to tree population
    clTreeSearch * p_oTreesToWrite; //pointer to search results
    clTree * p_oTree;
    FILE * out; //out file to write to
    char cTemp[100], //temp for string formatting
         cTempTreeVal[MAX_TREE_CHAR_VALUE_SIZE], //for getting char values
         cBuf[BUFFER_SIZE + 1]; //file writing buffer
    float fTemp; //for getting float values
    int iTimestep, //current timestep
         iNumTimesteps, //total number of timesteps
         iTemp, //for getting integer values
         i, j; //loop counters
    bool bTemp; //for getting bool values

    //Open the out file
    if ( strlen( m_cFileRoot ) == 0 ) return;

    out = fopen( cFilename, "at" );

    //If the file could not be opened - throw error
    if ( NULL == out )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_FILE;
      strcpy( stcErr.cFunction, "clOutputCFile::WriteTreeData" );
      sprintf( stcErr.cMoreInfo, "%s%s%s", "Could not open output file \"", cFilename, "\"" );
      throw( stcErr );
    }

    //Clear the buffer
    cBuf[0] = '\0';

    //Get the tree population
    p_oTemp = mp_oSimManager->GetPopulationObject( "treepopulation" );
    p_oTrees = ( clTreePopulation * ) p_oTemp;

    //Get the timestep data
    iTimestep = mp_oSimManager->GetCurrentTimestep();
    iNumTimesteps = mp_oSimManager->GetNumberOfTimesteps();

    //******************************
    // Write the tree map header
    //******************************
    if ( m_iNumTreeSettingsToSave > 0 )
    {
      strcpy( cBuf, "<tr_treemap>" );

      //Write a species list
      strcpy( cTemp, "<tm_speciesList>" );
      AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
      for ( i = 0; i < p_oTrees->GetNumberOfSpecies(); i++ )
      {
        sprintf( cTemp, "%s%s%s", "<tm_species speciesName=\"", p_oTrees->TranslateSpeciesCodeToName( i ), "\"/>" );
        AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
      }
      strcpy( cTemp, "</tm_speciesList>" );
      AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
    }
    //Tree settings for each species/type combo
    for ( i = 0; i < m_iNumTreeSettingsToSave; i++ )
    {
      //Tree settings opening tag
      sprintf( cTemp, "%s%s%s%d%s", "<tm_treeSettings sp=\"", p_oTrees->TranslateSpeciesCodeToName( mp_treeSettings[i].iSpecies ),
           "\" tp=\"", mp_treeSettings[i].iType, "\">" );

      AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );


      //Data member labels and codes
      if ( mp_treeSettings[i].iNumInts > 0 )
      { //ints

        strcpy( cTemp, "<tm_intCodes>" );
        AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

        for ( j = 0; j < mp_treeSettings[i].iNumInts; j++ )
        {
          sprintf( cTemp, "%s%s%s%d%s", "<tm_intCode label=\"",
               p_oTrees->GetIntDataLabel( mp_treeSettings[i].p_iIntCodes[j], mp_treeSettings[i].iSpecies, mp_treeSettings[i].iType ),
               "\">", mp_treeSettings[i].p_iIntCodes[j], "</tm_intCode>" );

          AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
        }

        strcpy( cTemp, "</tm_intCodes>" );
        AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

      } //end of if (mp_treeSettings[i].iNumInts > 0)
      if ( mp_treeSettings[i].iNumFloats > 0 )
      { //floats

        strcpy( cTemp, "<tm_floatCodes>" );
        AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

        for ( j = 0; j < mp_treeSettings[i].iNumFloats; j++ )
        {
          sprintf( cTemp, "%s%s%s%d%s", "<tm_floatCode label=\"",
               p_oTrees->GetFloatDataLabel( mp_treeSettings[i].p_iFloatCodes[j], mp_treeSettings[i].iSpecies, mp_treeSettings[i].iType ),
               "\">", mp_treeSettings[i].p_iFloatCodes[j], "</tm_floatCode>" );

          AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
        }

        strcpy( cTemp, "</tm_floatCodes>" );
        AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

      } //end of if (mp_treeSettings[i].iNumFloats > 0)

      if ( mp_treeSettings[i].iNumChars > 0 )
      { //chars

        strcpy( cTemp, "<tm_charCodes>" );
        AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

        for ( j = 0; j < mp_treeSettings[i].iNumChars; j++ )
        {

          sprintf( cTemp, "%s%s%s%d%s", "<tm_charCode label=\"",
               p_oTrees->GetCharDataLabel( mp_treeSettings[i].p_iCharCodes[j], mp_treeSettings[i].iSpecies, mp_treeSettings[i].iType ),
               "\">", mp_treeSettings[i].p_iCharCodes[j], "</tm_charCode>" );

          AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
        }
        strcpy( cTemp, "</tm_charCodes>" );
        AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
      } //end of if (mp_treeSettings[i].iNumChars > 0)

      if ( mp_treeSettings[i].iNumBools > 0 )
      { //bools

        strcpy( cTemp, "<tm_boolCodes>" );
        AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

        for ( j = 0; j < mp_treeSettings[i].iNumBools; j++ )
        {
          sprintf( cTemp, "%s%s%s%d%s", "<tm_boolCode label=\"",
               p_oTrees->GetBoolDataLabel( mp_treeSettings[i].p_iBoolCodes[j], mp_treeSettings[i].iSpecies, mp_treeSettings[i].iType ),
               "\">", mp_treeSettings[i].p_iBoolCodes[j], "</tm_boolCode>" );

          AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

        }

        strcpy( cTemp, "</tm_boolCodes>" );
        AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
      } //end of if (mp_treeSettings[i].iNumBools > 0)

      //Tree settings closing tag

      strcpy( cTemp, "</tm_treeSettings>" );
      AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
    } //end of for (i = 0; i < m_iNumTreeSettingsToSave; i++)

    //******************************
    // Tree data saving - go through the species/type combos and write the
    // trees for each
    //******************************
    for ( i = 0; i < m_iNumTreeSettingsToSave; i++ )
    {

      //Is this a saving timestep?
      //Increment the timestep counter and compare to frequency - always
      //write if this is timestep 0 or the last timestep
      mp_treeSettings[i].iSumTimestep++;
      if ( 0 == iTimestep || iNumTimesteps == iTimestep
           || ( mp_treeSettings[i].iSaveFreq == mp_treeSettings[i].iSumTimestep ) )
           {

             //Yes - it's time to save - assemble query string
             sprintf( cTemp, "%s%d%s%d", "species=", mp_treeSettings[i].iSpecies, "::type=", mp_treeSettings[i].iType );
             //Find the trees
             p_oTreesToWrite = p_oTrees->Find( cTemp );

             //Write the trees
             p_oTree = p_oTreesToWrite->NextTree();
             while ( p_oTree )
             {
               sprintf( cTemp, "%s%d%s%d%s", "<tree sp=\"", p_oTree->GetSpecies(), "\" tp=\"", p_oTree->GetType(), "\">" );
               AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

               //Write int values for this tree
               for ( j = 0; j < mp_treeSettings[i].iNumInts; j++ )
               {
                 p_oTree->GetValue( mp_treeSettings[i].p_iIntCodes[j], & iTemp );
                 sprintf( cTemp, "%s%d%s%d%s", "<int c=\"", mp_treeSettings[i].p_iIntCodes[j], "\">", iTemp, "</int>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
               }
               //Write float values for this tree
               for ( j = 0; j < mp_treeSettings[i].iNumFloats; j++ )
               {
                 p_oTree->GetValue( mp_treeSettings[i].p_iFloatCodes[j], & fTemp );

                 sprintf( cTemp, "%s%d%s%g%s", "<fl c=\"", mp_treeSettings[i].p_iFloatCodes[j], "\">", fTemp, "</fl>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
               }
               //Write char values for this tree
               for ( j = 0; j < mp_treeSettings[i].iNumChars; j++ )
               {
                 p_oTree->GetValue( mp_treeSettings[i].p_iCharCodes[j], cTempTreeVal );
                 if ( strlen( cTempTreeVal ) != 0 )
                 {
                   sprintf( cTemp, "%s%d%s%s%s", "<ch c=\"", mp_treeSettings[i].p_iCharCodes[j], "\">",
                        cTempTreeVal, "</ch>" );

                   AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                 }
               }
               //Write bool values for this tree
               for ( j = 0; j < mp_treeSettings[i].iNumBools; j++ )
               {
                 p_oTree->GetValue( mp_treeSettings[i].p_iBoolCodes[j], & bTemp );
                 if ( true == bTemp )
                   sprintf( cTemp, "%s%d%s", "<bl c=\"", mp_treeSettings[i].p_iBoolCodes[j], "\">true</bl>" );
                 else
                   sprintf( cTemp, "%s%d%s", "<bl c=\"", mp_treeSettings[i].p_iBoolCodes[j], "\">false</bl>" );

                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
               }

               strcpy( cTemp, "</tree>" );
               AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

               p_oTree = p_oTreesToWrite->NextTree();
             }

             //Reset the counter
             mp_treeSettings[i].iSumTimestep = 0;
      } //end of if (mp_treeSettings[i].iSumFreq == mp_treeSetti...
    } //end of for (i = 0; i < m_iNumTreeSettingsToSave; i++)

    //Write the end tag for the tree map
    if ( m_iNumTreeSettingsToSave > 0 )
    {

      //Flush the buffer one last time
      fwrite( cBuf, strlen( cBuf ), 1, out );
      strcpy( cBuf, "</tr_treemap>" );
      fwrite( cBuf, strlen( cBuf ), 1, out );

    }
    fclose( out );
  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::WriteTreeData" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
//     WriteTimestepHeader()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

/////////////////////////////////////////////////////////////////////////////
void clOutput::WriteTimestepHeader( const char * cFilename )
{
  try
  {
    FILE * out; //file writing stream
    char cTemp[50], cFileCode[9]; //file code
    int iFileVersion = 1; //for the filecode

    //If the filename is empty return
    if ( strlen( m_cFileRoot ) == 0 ) return;

    //Open the out file
    out = fopen( cFilename, "wt" );

    //If the file could not be opened - throw error
    if ( NULL == out )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_FILE;
      strcpy( stcErr.cFunction, "clOutput::WriteTimestepHeader" );
      sprintf( stcErr.cMoreInfo, "%s%s%s", "Could not open output file \"", cFilename, "\"" );
      throw( stcErr );
    }

    AssembleFileCode( detailed_output_timestep, iFileVersion, cFileCode );
    //***************************
    //Write the header of the file
    //***************************
    //Root element
    strcpy( cTemp, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" );
    fwrite( cTemp, strlen( cTemp ), 1, out );

    //  << "<!DOCTYPE timestepRundata SYSTEM \"xml\\RundataTimestep.dtd\">\n"
    sprintf( cTemp, "%s%s%s", "<timestepRundata fileCode=\"", cFileCode, "\">" );
    fwrite( cTemp, strlen( cTemp ), 1, out );

    //Timestep
    sprintf( cTemp, "%s%d%s", "<rt_timestep>", mp_oSimManager->GetCurrentTimestep(), "</rt_timestep>" );
    fwrite( cTemp, strlen( cTemp ), 1, out );

    fclose( out );
  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::WriteTimestepHeader" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
//     WriteTimestepFooter()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

/////////////////////////////////////////////////////////////////////////////
void clOutput::WriteTimestepFooter( const char * cFilename )
{
  try
  {
    char cCloser[] = "</timestepRundata>";
    FILE * out; //file writing stream

    //If the filename is empty return
    if ( strlen( cFilename ) == 0 ) return;

    //Open the out file
    out = fopen( cFilename, "at" );

    //If we can't open the file throw an error
    if ( NULL == out )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_FILE;
      strcpy( stcErr.cFunction, "clOutput::WriteTimestepFooter" );
      sprintf( stcErr.cMoreInfo, "%s%s%s", "Couldn't open file \"", cFilename, "\"." );
      throw( stcErr );
    }

    fwrite( cCloser, strlen( cCloser ), 1, out );
    fclose( out );

  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::WriteTimestepFooter" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
//     WriteGridData()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

/////////////////////////////////////////////////////////////////////////////
void clOutput::WriteGridData( const char * cFilename )
{
  try
  {
    char cTempTreeVal[MAX_TREE_CHAR_VALUE_SIZE], //for getting char values
         cBuf[BUFFER_SIZE + 1], //buffer for file writing
         cTemp[100]; //for formatting data into buffer
    clPlot * p_oPlot; //plot object - for grid lengths
    clPackage * p_oPackage; //for writing the grid's packages
    FILE * out; //file to write grid file
    float fPlotLenX, fPlotLenY, //plot dimensions
         fTemp; //for getting float values
    int iTimestep, //current timestep
         iNumTimesteps, //total number of timesteps
         iNumXCells, iNumYCells, //number of X and Y cells in the grid
         iNumDataMembers, //number of a kind of data member
         iTemp, //for getting int values
         i, j, iX, iY; //loop counters
    bool bSavePackages, //for whether or not packages are to be saved
         bTemp; //for getting bool values

    //If the filename is empty return
    if ( strlen( cFilename ) == 0 ) return;

    //Open the out file
    out = fopen( cFilename, "at" );

    //If the file could not be opened - throw error
    if ( NULL == out )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_FILE;
      strcpy( stcErr.cFunction, "clOutput::WriteGridData" );
      sprintf( stcErr.cMoreInfo, "%s%s%s", "Could not open output file \"", cFilename, "\"" );
      throw( stcErr );
    }

    //Get the timestep data
    iTimestep = mp_oSimManager->GetCurrentTimestep();
    iNumTimesteps = mp_oSimManager->GetNumberOfTimesteps();

    //Get the plot dimensions
    p_oPlot = mp_oSimManager->GetPlotObject();
    fPlotLenX = p_oPlot->GetXPlotLength();
    fPlotLenY = p_oPlot->GetYPlotLength();

    //Loop through the grids to be written
    for ( i = 0; i < m_iNumGridsToSave; i++ )
    {
      //Clear the buffer
      cBuf[0] = '\0';

      //Is this a saving timestep?
      //Increment the timestep counter and compare to frequency - always
      //write if this is timestep 0 or the last timestep
      mp_gridSettings[i].iSumTimestep++;
      if ( 0 == iTimestep || iNumTimesteps == iTimestep
           || ( mp_gridSettings[i].iSaveFreq == mp_gridSettings[i].iSumTimestep ) )
           {
             //Reset counter
             mp_gridSettings[i].iSumTimestep = 0;

             //Figure out if there are any package data members to save
             if ( mp_gridSettings[i].iNumPackageInts > 0 || mp_gridSettings[i].iNumPackageFloats > 0
                  || mp_gridSettings[i].iNumPackageChars > 0 || mp_gridSettings[i].iNumPackageBools > 0 )
                    bSavePackages = true;
             else
               bSavePackages = false;

             //********************************
             // Grid header
             //********************************

             //Open tag for grid
             sprintf( cTemp, "%s%s%s", "<grid gridName=\"", mp_gridSettings[i].p_oGridPointer->GetName(), "\">" );
             AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

             //Data labels - write all even if all are not being saved
             //Ints
             iNumDataMembers = mp_gridSettings[i].p_oGridPointer->GetNumberIntDataMembers();
             if ( iNumDataMembers > 0 )
             {
               strcpy( cTemp, "<ma_intCodes>" );
               AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

               for ( j = 0; j < iNumDataMembers; j++ )
               {
                 sprintf( cTemp, "%s%s%s%d%s", "<ma_intCode label=\"", mp_gridSettings[i].p_oGridPointer->GetIntDataLabel( j ),
                      "\">", j, "</ma_intCode>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
               }

               strcpy( cTemp, "</ma_intCodes>" );
               AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
             } //end of if (iNumDataMembers > 0)

             //floats
             iNumDataMembers = mp_gridSettings[i].p_oGridPointer->GetNumberFloatDataMembers();
             if ( iNumDataMembers > 0 )
             {
               strcpy( cTemp, "<ma_floatCodes>" );
               AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

               for ( j = 0; j < iNumDataMembers; j++ )
               {
                 sprintf( cTemp, "%s%s%s%d%s", "<ma_floatCode label=\"", mp_gridSettings[i].p_oGridPointer->GetFloatDataLabel( j ),
                      "\">", j, "</ma_floatCode>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
               }

               strcpy( cTemp, "</ma_floatCodes>" );
               AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
             } //end of if (iNumDataMembers > 0)

             //chars
             iNumDataMembers = mp_gridSettings[i].p_oGridPointer->GetNumberCharDataMembers();
             if ( iNumDataMembers > 0 )
             {
               strcpy( cTemp, "<ma_charCodes>" );
               AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

               for ( j = 0; j < iNumDataMembers; j++ )
               {
                 sprintf( cTemp, "%s%s%s%d%s", "<ma_charCode label=\"", mp_gridSettings[i].p_oGridPointer->GetCharDataLabel( j ),
                      "\">", j, "</ma_charCode>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
               }

               strcpy( cTemp, "</ma_charCodes>" );
               AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
             } //end of if (iNumDataMembers > 0)

             //bools
             iNumDataMembers = mp_gridSettings[i].p_oGridPointer->GetNumberBoolDataMembers();
             if ( iNumDataMembers > 0 )
             {
               strcpy( cTemp, "<ma_boolCodes>" );
               AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

               for ( j = 0; j < iNumDataMembers; j++ )
               {
                 sprintf( cTemp, "%s%s%s%d%s", "<ma_boolCode label=\"", mp_gridSettings[i].p_oGridPointer->GetBoolDataLabel( j ),
                      "\">", j, "</ma_boolCode>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
               }

               strcpy( cTemp, "</ma_boolCodes>" );
               AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
             } //end of if (iNumDataMembers > 0)

             //Package data labels
             if ( mp_gridSettings[i].p_oGridPointer->GetPackageDataChanged() )
             {
               //Ints
               iNumDataMembers = mp_gridSettings[i].p_oGridPointer->GetNumberIntPackageDataMembers();
               if ( iNumDataMembers > 0 )
               {
                 strcpy( cTemp, "<ma_packageIntCodes>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

                 for ( j = 0; j < iNumDataMembers; j++ )
                 {
                   sprintf( cTemp, "%s%s%s%d%s", "<ma_intCode label=\"", mp_gridSettings[i].p_oGridPointer->GetPackageIntDataLabel( j ),
                        "\">", j, "</ma_intCode>" );
                   AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                 }

                 strcpy( cTemp, "</ma_packageIntCodes>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
               } //end of if (iNumDataMembers > 0)

               //Floats
               iNumDataMembers = mp_gridSettings[i].p_oGridPointer->GetNumberFloatPackageDataMembers();
               if ( iNumDataMembers > 0 )
               {
                 strcpy( cTemp, "<ma_packageFloatCodes>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

                 for ( j = 0; j < iNumDataMembers; j++ )
                 {
                   sprintf( cTemp, "%s%s%s%d%s", "<ma_floatCode label=\"", mp_gridSettings[i].p_oGridPointer->GetPackageFloatDataLabel( j ),
                        "\">", j, "</ma_floatCode>" );
                   AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                 }

                 strcpy( cTemp, "</ma_packageFloatCodes>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
               } //end of if (iNumDataMembers > 0)

               //Chars
               iNumDataMembers = mp_gridSettings[i].p_oGridPointer->GetNumberCharPackageDataMembers();
               if ( iNumDataMembers > 0 )
               {
                 strcpy( cTemp, "<ma_packageCharCodes>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

                 for ( j = 0; j < iNumDataMembers; j++ )
                 {
                   sprintf( cTemp, "%s%s%s%d%s", "<ma_charCode label=\"", mp_gridSettings[i].p_oGridPointer->GetPackageCharDataLabel( j ),
                        "\">", j, "</ma_charCode>" );
                   AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                 }

                 strcpy( cTemp, "</ma_packageCharCodes>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
               } //end of if (iNumDataMembers > 0)

               //Bools
               iNumDataMembers = mp_gridSettings[i].p_oGridPointer->GetNumberBoolPackageDataMembers();
               if ( iNumDataMembers > 0 )
               {
                 strcpy( cTemp, "<ma_packageBoolCodes>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

                 for ( j = 0; j < iNumDataMembers; j++ )
                 {
                   sprintf( cTemp, "%s%s%s%d%s", "<ma_boolCode label=\"", mp_gridSettings[i].p_oGridPointer->GetPackageBoolDataLabel( j ),
                        "\">", j, "</ma_boolCode>" );
                   AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                 }

                 strcpy( cTemp, "</ma_packageBoolCodes>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
               } //end of if (iNumDataMembers > 0)
             }

             //Plot size values
             sprintf( cTemp, "%s%g%s%g%s", "<ma_plotLenX>", fPlotLenX, "</ma_plotLenX><ma_plotLenY>",
                  fPlotLenY, "</ma_plotLenY>" );
             AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

             //Length of grids
             sprintf( cTemp, "%s%g%s%g%s", "<ma_lengthXCells>", mp_gridSettings[i].p_oGridPointer->GetLengthXCells(),
                  "</ma_lengthXCells><ma_lengthYCells>", mp_gridSettings[i].p_oGridPointer->GetLengthYCells(),
                  "</ma_lengthYCells>" );
             AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

             //********************************
             // Map Values
             //********************************
             //Escape out if there's nothing to save
             if ( mp_gridSettings[i].iNumInts == 0 && mp_gridSettings[i].iNumFloats == 0 && mp_gridSettings[i].iNumChars == 0
                  && mp_gridSettings[i].iNumBools == 0 && mp_gridSettings[i].iNumPackageInts == 0
                  && mp_gridSettings[i].iNumPackageFloats == 0 && mp_gridSettings[i].iNumPackageChars == 0
                  && mp_gridSettings[i].iNumPackageBools == 0 )
                    goto EndTag;

             iNumXCells = mp_gridSettings[i].p_oGridPointer->GetNumberXCells();
             iNumYCells = mp_gridSettings[i].p_oGridPointer->GetNumberYCells();
             for ( iX = 0; iX < iNumXCells; iX++ )
               for ( iY = 0; iY < iNumYCells; iY++ )
               {

                 //opening grid tag
                 sprintf( cTemp, "%s%d%s%d%s", "<ma_v x=\"", iX, "\" y=\"", iY, "\">" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

                 //Write int values for this cell
                 for ( j = 0; j < mp_gridSettings[i].iNumInts; j++ )
                 {

                   mp_gridSettings[i].p_oGridPointer->GetValueOfCell( iX, iY, mp_gridSettings[i].p_iIntCodes[j], & iTemp );

                   sprintf( cTemp, "%s%d%s%d%s", "<int c=\"", mp_gridSettings[i].p_iIntCodes[j], "\">", iTemp, "</int>" );

                   AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                 }
                 //Write float values for this cell
                 for ( j = 0; j < mp_gridSettings[i].iNumFloats; j++ )
                 {

                   mp_gridSettings[i].p_oGridPointer->GetValueOfCell( iX, iY, mp_gridSettings[i].p_iFloatCodes[j], & fTemp );

                   sprintf( cTemp, "%s%d%s%g%s", "<fl c=\"", mp_gridSettings[i].p_iFloatCodes[j], "\">", fTemp, "</fl>" );

                   AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                 }
                 //Write char values for this cell - but don't write empty strings
                 //map reading chokes on empty strings
                 for ( j = 0; j < mp_gridSettings[i].iNumChars; j++ )
                 {

                   mp_gridSettings[i].p_oGridPointer->GetValueOfCell( iX, iY,
                        mp_gridSettings[i].p_iCharCodes[j], cTempTreeVal );

                   if ( strlen( cTempTreeVal ) > 0 )
                   {
                     sprintf( cTemp, "%s%d%s%s%s", "<ch c=\"", mp_gridSettings[i].p_iCharCodes[j], "\">",
                          cTempTreeVal, "</ch>" );

                     AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                   }
                 }
                 //Write bool values for this cell
                 for ( j = 0; j < mp_gridSettings[i].iNumBools; j++ )
                 {

                   mp_gridSettings[i].p_oGridPointer->GetValueOfCell( iX, iY, mp_gridSettings[i].p_iBoolCodes[j], & bTemp );

                   if ( true == bTemp )
                     sprintf( cTemp, "%s%d%s", "<bl c=\"", mp_gridSettings[i].p_iBoolCodes[j], "\">true</bl>" );

                   else
                     sprintf( cTemp, "%s%d%s", "<bl c=\"", mp_gridSettings[i].p_iBoolCodes[j], "\">false</bl>" );

                   AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                 }

                 //Packages
                 if ( bSavePackages )
                 {
                   p_oPackage = mp_gridSettings[i].p_oGridPointer->GetFirstPackageOfCell( iX, iY );
                   while ( p_oPackage )
                   {

                     strcpy( cTemp, "<pkg>" );
                     AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

                     //Write int values for this package
                     for ( j = 0; j < mp_gridSettings[i].iNumPackageInts; j++ )
                     {

                       p_oPackage->GetValue( mp_gridSettings[i].p_iPackageIntCodes[j], & iTemp );

                       sprintf( cTemp, "%s%d%s%d%s", "<pint c=\"", mp_gridSettings[i].p_iPackageIntCodes[j],
                            "\">", iTemp, "</pint>" );

                       AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                     }
                     //Write float values for this package
                     for ( j = 0; j < mp_gridSettings[i].iNumPackageFloats; j++ )
                     {

                       p_oPackage->GetValue( mp_gridSettings[i].p_iPackageFloatCodes[j], & fTemp );

                       sprintf( cTemp, "%s%d%s%g%s", "<pfl c=\"", mp_gridSettings[i].p_iPackageFloatCodes[j],
                            "\">", fTemp, "</pfl>" );

                       AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                     }
                     //Write char values for this package
                     for ( j = 0; j < mp_gridSettings[i].iNumPackageChars; j++ )
                     {

                       p_oPackage->GetValue( mp_gridSettings[i].p_iPackageCharCodes[j], cTempTreeVal );

                       if ( strlen( cTempTreeVal ) > 0 )
                       {

                         sprintf( cTemp, "<pch c=\"", mp_gridSettings[i].p_iPackageCharCodes[j], "\">",
                              cTempTreeVal, "</pch>" );

                         AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                       }
                     }
                     //Write bool values for this package
                     for ( j = 0; j < mp_gridSettings[i].iNumPackageBools; j++ )
                     {

                       p_oPackage->GetValue( mp_gridSettings[i].p_iPackageBoolCodes[j], & bTemp );

                       if ( true == bTemp )
                         sprintf( cTemp, "%s%d%s", "<pbl c=\"", mp_gridSettings[i].p_iPackageBoolCodes[j], "\">true</pbl>" );
                       else
                         sprintf( cTemp, "%s%d%s", "<pbl c=\"", mp_gridSettings[i].p_iPackageBoolCodes[j], "\">false</pbl>" );

                       AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
                     }
                     strcpy( cTemp, "</pkg>" );
                     AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );

                     p_oPackage = p_oPackage->GetNextPackage();
                   } //end of while (p_oPackage)
                 } //end of if (bSavePackages)

                 //closing value tag
                 strcpy( cTemp, "</ma_v>" );
                 AddToBuffer( cBuf, cTemp, out, BUFFER_SIZE );
               }
             //closing grid tag
           EndTag:
             //Flush buffer for the last time
             fwrite( cBuf, strlen( cBuf ), 1, out );
             strcpy( cBuf, "</grid>" );
             fwrite( cBuf, strlen( cBuf ), 1, out );
      } //end of if this is a saving timestep
    } //end of for (i = 0; i < m_iNumGridsToSave; i++)

    fclose( out );
  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::WriteGridData" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
//GetTimestepFilename()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)

/////////////////////////////////////////////////////////////////////////////
void clOutput::GetTimestepFilename( int iTimestep, char * cFilename )
{
  try
  {
    //If the filename is empty return empty string
    if ( strlen( m_cFileRoot ) == 0 )
    {
      //Empty the filename string passed
      cFilename[0] = '\0';
      return;
    }

    sprintf( cFilename, "%s%s%d%s", m_cFileRoot, "_", iTimestep, DETAILED_OUTPUT_FILE_EXT );

  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clOutput::GetTimestepFilename" );
    throw( stcErr );
  }
}
//---------------------------------------------------------------------------
