//---------------------------------------------------------------------------
// SimpleLinearGrowth.cpp
//---------------------------------------------------------------------------
#include "SimpleLinearGrowth.h"
#include "TreePopulation.h"
#include "SimManager.h"
#include "ParsingFunctions.h"
#include "GrowthOrg.h"
#include "ModelMath.h"

//////////////////////////////////////////////////////////////////////////////
// Constructor
// Edit history:
// -------------
// June 10, 2004 - Created (LEM)
/////////////////////////////////////////////////////////////////////////////*/
clSimpleLinearGrowth::clSimpleLinearGrowth( clSimManager * p_oSimManager ) : clGrowthBase( p_oSimManager ), clBehaviorBase( p_oSimManager ),
     clWorkerBase( p_oSimManager )
     {

       try
       {
         mp_fSlope = NULL;
         mp_fIntercept = NULL;

         strcpy( m_cNameString, "simplelineargrowthshell" );
       }
       catch ( modelErr & err )
       {
         throw( err );
       }
       catch ( modelMsg & msg )
       {
         throw( msg );
       } //non-fatal error
       catch ( ... )
       {
         modelErr stcErr;
         stcErr.iErrorCode = UNKNOWN;
         strcpy( stcErr.cFunction, "clSimpleLinearGrowth::clSimpleLinearGrowth" );
         throw( stcErr );
       }
}

//---------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
// Destructor
// Edit history:
// -------------
// June 10, 2004 - Created (LEM)
/////////////////////////////////////////////////////////////////////////////*/
clSimpleLinearGrowth::~clSimpleLinearGrowth()
{

  delete[] mp_fSlope;
  delete[] mp_fIntercept;
}

//---------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
// DoShellSetup()
// Edit history:
// -------------
// June 10, 2004 - Created (LEM)
// August 27, 2004 - Made light required (LEM)
/////////////////////////////////////////////////////////////////////////////*/
void clSimpleLinearGrowth::DoShellSetup( DOMDocument * p_oDoc )
{
  try
  {
    clPopulationBase * p_oTemp = mp_oSimManager->GetPopulationObject( "treepopulation" );
    clTreePopulation * p_oPop = ( clTreePopulation * ) p_oTemp;
    DOMElement * p_oElement = p_oDoc->getDocumentElement();
    floatVal * p_fTempValues; //for getting species-specific values
    short int iNumSpecies = p_oPop->GetNumberOfSpecies(), i;

    //Conversion factor
    if ( height_only == m_iGrowthMethod )
    {
      //The conversion factor needs to convert from cm to m and from
      //annual to timestep
      m_fConversionFactor = mp_oSimManager->GetNumberOfYearsPerTimestep() / 100;
    } else {
      //The conversion factor needs to convert from mm to cm, from annual to
      //timestep, and from radial to diameter
      m_fConversionFactor = mp_oSimManager->GetNumberOfYearsPerTimestep() / 10 * 2;
    }

    //Declare the arrays we'd like read
    mp_fSlope = new float[iNumSpecies];
    mp_fIntercept = new float[iNumSpecies];

    //Declare the species-specific temp array and pre-load with the species that
    //this behavior affects
    p_fTempValues = new floatVal[m_iNumBehaviorSpecies];
    for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
      p_fTempValues[i].code = mp_iWhatSpecies[i];

    if ( height_only == m_iGrowthMethod )
    {
      //Growth intercept
      FillSpeciesSpecificValue( p_oElement, "gr_heightSimpleLinearIntercept", "gr_hsliVal", p_fTempValues,
           m_iNumBehaviorSpecies, p_oPop, true );

      //Transfer values to our permanent array
      for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
        mp_fIntercept[p_fTempValues[i].code] = p_fTempValues[i].val;

      //Growth slope
      FillSpeciesSpecificValue( p_oElement, "gr_heightSimpleLinearSlope", "gr_hslsVal", p_fTempValues,
           m_iNumBehaviorSpecies, p_oPop, true );

      //Transfer values to our permanent array
      for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
        mp_fSlope[p_fTempValues[i].code] = p_fTempValues[i].val;
    }
    else
    {
      //Growth intercept
      FillSpeciesSpecificValue( p_oElement, "gr_diamSimpleLinearIntercept", "gr_dsliVal", p_fTempValues,
           m_iNumBehaviorSpecies, p_oPop, true );

      //Transfer values to our permanent array
      for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
        mp_fIntercept[p_fTempValues[i].code] = p_fTempValues[i].val;

      //Growth slope
      FillSpeciesSpecificValue( p_oElement, "gr_diamSimpleLinearSlope", "gr_dslsVal", p_fTempValues,
           m_iNumBehaviorSpecies, p_oPop, true );

      //Transfer values to our permanent array
      for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
        mp_fSlope[p_fTempValues[i].code] = p_fTempValues[i].val;
    }

    //Make sure all species/type combos have "Light" registered
    for (i = 0; i < m_iNumSpeciesTypeCombos; i++) {
      if (-1 == mp_oGrowthOrg->GetLightCode(mp_whatSpeciesTypeCombos[i].iSpecies,
                                      mp_whatSpeciesTypeCombos[i].iType)) {
        modelErr stcErr;
        strcpy( stcErr.cFunction, "clSimpleLinearGrowth::DoShellSetup" );
        sprintf( stcErr.cMoreInfo,
                "%s%d%s%d%s%s", "Type/species combo species=",
                mp_whatSpeciesTypeCombos[i].iSpecies,
                " type=", mp_whatSpeciesTypeCombos[i].iType,
                " does not have a required light behavior." );
                stcErr.iErrorCode = BAD_DATA;
        throw( stcErr );

      }
    }

    delete[] p_fTempValues;
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clSimpleLinearGrowth::DoShellSetup" );
    throw( stcErr );
  }
}
//---------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
// SetNameData()
// Edit history:
// -------------
// June 10, 2004 - Created (LEM)
/////////////////////////////////////////////////////////////////////////////*/
void clSimpleLinearGrowth::SetNameData(char *cNameString) {
 try {

   //Check the string passed and set the flags accordingly
   if (strcmp("simple linear growth", cNameString) == 0) {
     m_iGrowthMethod = diameter_auto;
   } else if (strcmp("simple linear growth diam only", cNameString) == 0) {
     m_iGrowthMethod = diameter_only;
   } else if (strcmp("simple linear growth height only", cNameString) == 0) {
     m_iGrowthMethod = height_only;
   }
   else {
     modelErr stcErr;
     stcErr.iErrorCode = BAD_DATA;
     sprintf(stcErr.cMoreInfo, "%s%s%s", "Unrecognized behavior name \"",
        cNameString, "\".");
     strcpy(stcErr.cFunction, "clSimpleLinearGrowth::SetNameData");
     throw(stcErr);
   }
 }
 catch (modelErr &err) {throw(err);}
 catch (modelMsg &msg) {throw(msg);} //non-fatal error
 catch (...) {
   modelErr stcErr;
   stcErr.iErrorCode = UNKNOWN;
   strcpy(stcErr.cFunction, "clSimpleLinearGrowth::SetNameData");
   throw(stcErr);
 }
}
//---------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
// CalculateFunctionValue()
// Edit history:
// -------------
// June 10, 2004 - Created (LEM)
/////////////////////////////////////////////////////////////////////////////*/
float clSimpleLinearGrowth::CalculateFunctionValue(clTree *p_oTree) {
 try {

  float fGli;
  short int iSpecies = p_oTree->GetSpecies(),
            iLightCode = mp_oGrowthOrg->GetLightCode(iSpecies, p_oTree->GetType());

   //Get the gli value for this tree
   p_oTree->GetValue(iLightCode, &fGli);

   //Calculate the function value
   return clModelMath::CalcPointValue(fGli, mp_fSlope[iSpecies], mp_fIntercept[iSpecies]);
 }
 catch (modelErr &err) {throw(err);}
 catch (modelMsg &msg) {throw(msg);} //non-fatal error
 catch (...) {
   modelErr stcErr;
   stcErr.iErrorCode = UNKNOWN;
   strcpy(stcErr.cFunction, "clSimpleLinearGrowth::CalculateFunctionValue");
   throw(stcErr);
 }
}
//---------------------------------------------------------------------------
