//---------------------------------------------------------------------------
// Storm
//---------------------------------------------------------------------------
#if !defined(Storm_H)
  #define Storm_H

#include "BehaviorBase.h"
#include "ModelMath.h"

class clGridBase;
class xercesc::DOMDocument;

/**
* Storms version 1.0
* The storm damage calculators function is to assess whether or not one or
* more storms has occurred, and if so, what the damage pattern and amount of
* damage were.   It creates a map of storm damage across the plot, with each
* location getting a damage index between 0 (no damage) and 1 (most damage).
*
* <b>Deciding when storms occur</b><br>
* The 0  1 interval of storm severity values is subdivided into ten storm
* severity classes.  Each class is assigned a return interval in the parameter
* file by the user.  Taking the reciprocal of the return interval gives us the
* annual probability of each type of storm.
*
* To decide whether a given storm occurs, a random number is compared to the
* annual probability.  Each storm severity class is decided individually.
* For multi-year timesteps, there is one random coin flip per year per
* severity class.  This allows multiple storms to occur in the same timestep.
* For multi-year timesteps, multiple storms of the same severity class can
* occur, up to one per year.  Storm damage for multiple storms is additive, up
* to a max damage index of 1.
*
* <b>Calculating damage when storms occur</b><br>
* The amount of damage caused by a timestep's storms is stored in the grid
* named "Storm Damage".  Each cell in the grid contains a value between 0 (no
* damage) and 1 (total damage).
*
* There are four damage patterns, and the value in the "Storm Damage" cells
* is calculated differently for each:
* <ol>
* <li><i>Varying spatial distribution, deterministic.</i>  In this pattern, a
* static storm vulnerability map is supplied for the plot in the parameter
* file.  The map is a set of values from 0 (no susceptibility) to 1 (maximum
* susceptibility).  The susceptibility value at a location is multiplied by the
* damage index of any storm that hits to arrive at a final damage index for
* that location.</li>
* <li><i>Varying spatial distribution, stochastic.</i>  In this pattern, a
* static storm vulnerability map is supplied for the plot in the parameter
* file.  The map is a set of values from 0 (no susceptibility) to 1 (maximum
* susceptibility).  The damage index of a storm is determined as for the
* Uniform stochastic method.  Then the susceptibility value at a location is
* multiplied by this damage index to arrive at a final damage index for that
* location.</li>
* <li><i>Uniform deterministic.</i>  All plot locations receive the storms
* designated damage index.</li>
* <li><i>Uniform stochastic.</i>  The storms designated severity is used as the
* mean in a probability distribution function to arrive at a final damage
* index.</li>
* </ol>
*
* This behavior's call string and name string are both "storm".
*
* Copyright 2004 Charles D. Canham
* @author Lora E. Murphy
* <br>Edit history:
* <br>-----------------
* <br>September 9, 2004 - Created (LEM)
*/
class clStorm : virtual public clBehaviorBase {

  /**This is for testing purposes - whenever there's a heavy stochastic
  * component to the operation of a behavior the only effective testing happens
  * when the testing class can get in and rig events to happen*/
  friend class clTestStorm;

  public:

  /**
  * Constructor.
  * @param p_oSimManager Sim Manager object.
  */
  clStorm(clSimManager *p_oSimManager);

  /**
  * Destructor.  Frees memory.
  */
  ~clStorm();

  /**
  * Does behavior setup.  It does the following:
  * <ol>
  * <li>Calls ReadParFile() to read parameter file values</li>
  * <li>Calls DoGridSetup() to set up the grids</li>
  * <li>Calls CalculateStormProbabilities() to calculate the annual storm
  * probabilities</li>
  * <li>Calls SetStochFuncPointer() to set the stochastic function pointer,
  * if appropriate</li>
  * </ol>
  * @param p_oDoc DOM tree of parsed input file.
  */
  void GetData(xercesc::DOMDocument *p_oDoc);

  /**
  * Resets all values in the "Storm Damage" grid to 0.
  */
  void TimestepCleanup();

  /**
  * Performs the storm calculations.  It loops through all of the storm
  * severity classes and, for each class and for each year per timestep, uses a
  * random number against the values in mp_fStormProbabilities to determine if
  * a storm of a given severity has occurred.  If a storm happens,
  * ApplyDamage() is called.
  */
  void Action();

  protected:

  /**
  * Storms grid
  * This grid is named "Storm Damage" and it contains the severity of all
  * storms that happened in the previous timestep.  It is of user-settable
  * resolution.  It contains one float data member, "dmg_index", which contains
  * the damage index as a value from 0 (no damage) to 1 (total damage).
  *
  * If there is a map for the "Storm Susceptibility" grid, and no other grid
  * resolution information for this grid, then this grid will use the resolution
  * of "Storm Susceptibility".
  */
  clGridBase *mp_oStormGrid;

  /**
  * Storm susceptibility map
  * This grid is named "Storm Susceptibility".  It has one float data member
  * called "index".  The value in "index" is a value from 0 (no storm
  * susceptibility) to 1 (total storm susceptibility).
  *
  * If the susceptibility = the enum value "mapped", then a grid map for this
  * grid is expected in the parameter file.  The values it contains will not
  * be changed throughout the run.  The grid resolution must match that of
  * "Storm Damage" grid.
  *
  * If the damage pattern is something other than "mapped", then this pointer
  * is left as NULL and the grid is not initialized or used.
  */
  clGridBase *mp_oSusceptibilityMap;

  /**Function pointer for the appropriate RandomDraw function*/
  float (clStorm::*RandomDraw)(const float &fNumber);

  /**Timestep probability of the occurence of a storm of each return interval.
  * Array size is m_iNumSeverityClasses.  These values are calculated in
  * CalculateStormProbabilities().*/
  float *mp_fStormProbabilities;

  /**Standard deviation for normal or lognormal probability distribution
  * functions.  This is only used if m_iStochasticity = stochastic and
  * m_iDistribution = lognormal or normal.  Value comes from the parameter
  * file.*/
  float mp_fStdDev;

  /**Clumping parameter for negative binomial probability distribution
  * function.  This is only used if m_iStochasticity = stochastic and
  * m_iDistribution = negative_binomial.  Value comes from the parameter
  * file.*/
  float mp_fClump;

  /**The total number of storm return intervals.  Hardcoded.*/
  int m_iNumSeverityClasses;
  /**Return code for the "dmg_index" data member of the "Storm Damage" grid*/
  int m_iDmgIndexCode;
  /**Return code for the "index" data member of the "Storm Susceptibility" grid*/
  int m_iSusceptIndexCode;

  /**Enum for describing the damage pattern.  This controls the method used
  * to calculate final damage indexes.*/
  enum susceptibility {mapped,  /**<Susceptibility is entered via a map*/
                       uniform  /**<Susceptibility is uniform across the plot*/
                      }
  m_iSusceptibility; /**<Variable holding the susceptibility pattern value.
  Value comes from parameter file.  If the susceptibility is uniform, all
  locations have a susceptibility index of 1.  If it is mapped, the
  "Storm Susceptibility" grid holds each location's susceptibility between
  0 and 1.*/

  /**Enum for stochasticity.*/
  enum stochasticity {deterministic, /**<Deterministic damage index calculations.*/
                   stochastic     /**<Stochastic damage index calculations.*/
                  }
  m_iStochasticity; /**<Variable holding the stochasticity. Value comes from
  the parameter file.  This controls how the damage pattern is applied -
  whether the same damage index is applied to all cells or whether it is
  randomized for each.*/

  /**Enum for listing probability distribution functions.  Poisson is
  * not included because it creates integers only - not appropriate.*/
  enum distribution_func {lognormal, /**<Lognormal distribution*/
                       normal, /**<Normal distribution*/
                       negative_binomial /**<Negative binomial distribution*/
                      } m_iDistribution; /**<What probability distribution
                      function to use if the value in m_iDamagePattern is
                      "stochastic".  Value comes from the parameter file.*/

  /**
  * Does the grid setup for the behavior.  Steps:
  * <ol>
  * <li>Check to see if m_iSusceptibility is "mapped".  If it is, get the
  * "Storm Susceptibility" grid and verify that it has been created and that all
  * values in it are between 0 and 1.</li>
  * <li>Check for the existence of the "Storm Damage" grid.  If it has been
  * created in the parameter file, make sure the grid cell resolution matches
  * "Storm Susceptibility", if that grid exists.</li>
  * <li>If "Storm Damage" was not created in the parameter file, create it with
  * "Storm Susceptibility"'s grid cell resolution if that grid exists, or the
  * default grid cell resolution if it doesn't.</li>
  * <li>Call TimestepCleanup() to make sure that all values in "Storm Damage"
  * are 0.</li>
  * </ol>
  *
  * @throws modelErr if:
  * <ul>
  * <li>The damage pattern as read from the parameter file is "mapped" but there
  * is no map in the parameter file for the grid "Storm Susceptibility"</li>
  * <li>Either grid is in the parameter file but not set up correctly</li>
  * <li>There is grid cell resolution data in the parameter file for both
  * "Storm Damage" and "Storm Susceptibility" and they do not match</li>
  * </ul>
  */
  void DoGridSetup();

  /**
  * Calculates the annual probability of a storm of each severity class.
  * The probability of a storm of a given return interval is calculated as
  * 1/length of the return interval.  The return interval values should already
  * be in mp_fStormProbabilities, put there by ReadParFile().
  */
  void CalculateStormProbabilities();

  /**
  * Applies the damage for a storm of a given return interval.  The storm's
  * mean severity is calculated by taking a random draw on its severity
  * interval.  The damage is applied according to one of the following
  * scenarios:
  * <ul>
  * <li>If the storm stochasticity is deterministic and the susceptibility is
  * uniform, the value of the storm damage index is added to each cell of the
  * "Storm Damage" grid (up to a max of 1).</li>
  * <li>If the storm stochasticity is stochastic and the susceptibility is
  * uniform, for each cell in the "Storm Damage" grid, the value of the storm
  * damage index is fed as the mean to the appropriate random draw function in
  * the math library to get a new storm damage index; then this value is added
  * to the cell (up to a max of 1).</li>
  * <li>If the storm stochasticity is deterministic and the susceptibility is
  * mapped, the storm damage index is multiplied by the appropriate cell value
  * in the "Storm Susceptibility" grid and added to the cell of the "Storm
  * Damage" grid (up to a max of 1).</li>
  * <li>If the storm stochasticity is stochastic and the susceptibility is
  * mapped, for each cell in the "Storm Damage" grid, the value of the storm
  * damage index is fed as the mean to the appropriate random draw function in
  * the math library to get a new storm damage index.  This index is multiplied
  * by the appropriate cell value in the "Storm Susceptibility" grid and added
  * to the cell of the "Storm Damage" grid (up to a max of 1).</li>
  * </ul>
  * @param iSeverityClass Index of the severity class, up to
  * m_iNumSeverityClasses.  This is NOT the actual number of years in the
  * return interval.  This is not validated to ensure that it is a proper array
  * index before being used as such.
  */
  void ApplyDamage(int iSeverityClass);

  /**
  * Reads in parameters from the parameter file.
  * @throws modelErr if:
  * <ul>
  * <li>Any return interval is not greater than 0.</li>
  * <li>The susceptibility value is not recognized.</li>
  * <li>The stochasticity value is not recognized.</li>
  * <li>The stochasticity is "stochastic" and there is no value for
  * probability distribution function, or that value is not recognized.</li>
  * <li>The stochasticity is "stochastic", the probability distribution
  * function is "normal" or "lognormal", and there is no value for standard
  * deviation.</li>
  * <li>The stochasticity is "stochastic", the probability distribution
  * function is "negative_binomial", and there is no value for the clumping
  * parameter.</li>
  * </ul>
  * @param p_oDoc DOM tree of parsed input file.
  */
  void ReadParFile(xercesc::DOMDocument *p_oDoc);

  /**
  * Sets the function pointer for a stochastic damage pattern.  If
  * m_iDamagePattern is not "stochastic", nothing happens.  If it is, this sets
  * the value for clStorm::*RandomDraw depending on the value in
  * m_iDistribution.
  */
  void SetStochFuncPointer();

  /**
  * Performs a random draw on a normal distribution.  The standard deviation
  * is mp_fStdDev.
  * @param fMean Mean of the normal distribution.
  * @return Random number between 0 and 1.
  */
  inline float NormalDraw(const float &fMean) {
    float fVal = fMean + clModelMath::NormalRandomDraw(mp_fStdDev);
    if (fVal < 0) fVal = 0; if (fVal > 1) fVal = 1;
    return fVal;
  };

  /**
  * Performs a random draw on a lognormal distribution.  The standard deviation
  * is mp_fStdDev.
  * @param fMean Mean of the lognormal distribution.
  * @return Random number between 0 and 1.
  */
  inline float LognormalDraw(const float &fMean) {
    return clModelMath::LognormalRandomDraw(fMean, mp_fStdDev);
  };

  /**
  * Performs a random draw on a negative binomial distribution.  The clumping
  * parameter is mp_fClump.
  * @param fMean Mean of the negative binomial distribution.
  * @return Random number between 0 and 1.
  */
  inline float NegBinomialDraw(const float &fMean) {
    return clModelMath::NegBinomialRandomDraw(fMean, mp_fClump);
  };

};
//---------------------------------------------------------------------------
#endif // Storm_H
