//---------------------------------------------------------------------------
// StormDamageApplier.cpp
//---------------------------------------------------------------------------
#include "StormDamageApplier.h"
#include "TreePopulation.h"
#include "SimManager.h"
#include "ParsingFunctions.h"
#include "GridBase.h"
#include "ModelMath.h"
#include <math.h>

//////////////////////////////////////////////////////////////////////////////
// Constructor
// Edit history:
// -------------
// September 29, 2004 - Created (LEM)
// April 1, 2005 - Added stuff so this can kill trees (LEM)
// April 28, 2005 - Added new tree-killing stuff (LEM)
//////////////////////////////////////////////////////////////////////////////
clStormDamageApplier::clStormDamageApplier( clSimManager * p_oSimManager ) : clBehaviorBase( p_oSimManager ),
     clWorkerBase( p_oSimManager )
     {
       try
       {
         //Set namestring
         strcpy( m_cNameString, "storm damage applier" );

         m_cQuery = NULL;
         mp_fMinStormDBH = NULL;
         mp_fStmDmgInterceptMed = NULL;
         mp_fStmDmgInterceptFull = NULL;
         mp_fStmIntensityCoeff = NULL;
         mp_fStmDBHCoeff = NULL;
         mp_iStmDmgCodes = NULL;
         mp_iDeadCodes = NULL; //LEM 04/28/05
         mp_fPropDieMedDmg = NULL; //LEM 04/01/05
         mp_fPropDieFullDmg = NULL; //LEM 04/01/05
         mp_fPropTipUp = NULL; //LEM 04/28/05

         //Set that we want to add one int value
         m_iNewTreeInts = 1;

         //Allowed file types
         m_iNumAllowedTypes = 2;
         mp_iAllowedFileTypes = new int[m_iNumAllowedTypes];
         mp_iAllowedFileTypes[0] = parfile;
         mp_iAllowedFileTypes[1] = detailed_output;

         //Versions
         m_fVersionNumber = 2.0;
         m_fMinimumVersionNumber = 2.0;

       }
       catch ( modelErr & err )
       {
         throw( err );
       }
       catch ( modelMsg & msg )
       {
         throw( msg );
       } //non-fatal error
       catch ( ... )
       {
         modelErr stcErr;
         stcErr.iErrorCode = UNKNOWN;
         strcpy( stcErr.cFunction, "clStormDamageApplier::clStormDamageApplier" );
         throw( stcErr );
       }
}

////////////////////////////////////////////////////////////////////////////
// Destructor
// Edit history:
// -------------
// September 29, 2004 - Created (LEM)
// April 1, 2005 - Added stuff so this can kill trees (LEM)
// April 28, 2005 - Added new tree-killing stuff (LEM)
////////////////////////////////////////////////////////////////////////////
clStormDamageApplier::~clStormDamageApplier()
{
  delete[] m_cQuery;
  delete[] mp_fMinStormDBH;
  delete[] mp_fStmDmgInterceptMed;
  delete[] mp_fStmDmgInterceptFull;
  delete[] mp_fStmIntensityCoeff;
  delete[] mp_fStmDBHCoeff;
  delete[] mp_fPropDieMedDmg; //LEM 04/01/05
  delete[] mp_fPropDieFullDmg; //LEM 04/01/05
  delete[] mp_fPropTipUp; //LEM 04/28/05
  if ( mp_iStmDmgCodes )
  {
    for ( int i = 0; i < m_iNumTypes; i++ )
      delete[] mp_iStmDmgCodes[i];
  }
  delete[] mp_iStmDmgCodes;
  if ( mp_iDeadCodes ) //LEM 04/28/05
  {
    for ( int i = 0; i < m_iNumTypes; i++ )
      delete[] mp_iDeadCodes[i];
  }
  delete[] mp_iDeadCodes;
}

////////////////////////////////////////////////////////////////////////////
// RegisterTreeDataMembers()
// Edit history:
// -------------
// September 29, 2004 - Created (LEM)
////////////////////////////////////////////////////////////////////////////
void clStormDamageApplier::RegisterTreeDataMembers()
{
  clPopulationBase * p_oTemp = mp_oSimManager->GetPopulationObject( "treepopulation" );
  clTreePopulation * p_oPop = ( clTreePopulation * ) p_oTemp;
  int i, j, iNumSpecies = p_oPop->GetNumberOfSpecies();
  m_iNumTypes = p_oPop->GetNumberOfTypes();

  mp_iStmDmgCodes = new int * [m_iNumTypes];
  for ( i = 0; i < m_iNumTypes; i++ )
  {
    mp_iStmDmgCodes[i] = new int[iNumSpecies];
    for (j = 0; j < iNumSpecies; j++) {
      mp_iStmDmgCodes[i][j] = -1;
    }
  }

  for ( i = 0; i < m_iNumSpeciesTypeCombos; i++ )
  {
    if ( clTreePopulation::seedling == mp_whatSpeciesTypeCombos[i].iType )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cFunction, "clStormDamageApplier::RegisterTreeDataMembers" );
      strcpy( stcErr.cMoreInfo, "The storm damage applier behavior cannot be applied to seedlings." );
      throw( stcErr );
    }

    //Register the code and capture it
    mp_iStmDmgCodes[mp_whatSpeciesTypeCombos[i].iType] [mp_whatSpeciesTypeCombos[i].iSpecies] =
         p_oPop->RegisterInt( "stm_dmg", mp_whatSpeciesTypeCombos[i].iSpecies, mp_whatSpeciesTypeCombos[i].iType );
  }
}

////////////////////////////////////////////////////////////////////////////
// GetDeadCodes()
// Edit history:
// -------------
// April 28, 2005 - Created (LEM)
////////////////////////////////////////////////////////////////////////////
void clStormDamageApplier::GetDeadCodes(clTreePopulation *p_oPop)
{
  int i, iNumSpecies = p_oPop->GetNumberOfSpecies();

  mp_iDeadCodes = new int * [m_iNumTypes];
  for ( i = 0; i < m_iNumTypes; i++ )
  {
    mp_iDeadCodes[i] = new int[iNumSpecies];
  }

  for ( i = 0; i < m_iNumSpeciesTypeCombos; i++ )
  {
    //Get the code
    mp_iDeadCodes[mp_whatSpeciesTypeCombos[i].iType] [mp_whatSpeciesTypeCombos[i].iSpecies] =
         p_oPop->GetBoolDataCode( "dead", mp_whatSpeciesTypeCombos[i].iSpecies, mp_whatSpeciesTypeCombos[i].iType );
    if (-1 == mp_iDeadCodes[mp_whatSpeciesTypeCombos[i].iType] [mp_whatSpeciesTypeCombos[i].iSpecies]) {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cFunction, "clStormDamageApplier::GetDeadCodes" );
      strcpy( stcErr.cMoreInfo, "You must use tree mortality behaviors along with the storm damage applier." );
      throw( stcErr );
    }
  }
}

////////////////////////////////////////////////////////////////////////////
// GetParameterFileData()
// Edit history:
// -------------
// September 29, 2004 - Created (LEM)
// April 1, 2005 - Added stuff so this can kill trees (LEM)
// April 28, 2005 - Added new tree-killing stuff (LEM)
////////////////////////////////////////////////////////////////////////////
void clStormDamageApplier::GetParameterFileData( xercesc::DOMDocument * p_oDoc )
{
  clTreePopulation * p_oPop = ( clTreePopulation * ) mp_oSimManager->GetPopulationObject( "treepopulation" );
  DOMElement * p_oElement = p_oDoc->getDocumentElement();
  floatVal * p_fTempValues; //for getting species-specific values
  int i, iNumSpecies = p_oPop->GetNumberOfSpecies();

  //Declare our arrays
  mp_fMinStormDBH = new float[iNumSpecies];
  mp_fStmDmgInterceptMed = new float[iNumSpecies];
  mp_fStmDmgInterceptFull = new float[iNumSpecies];
  mp_fStmIntensityCoeff = new float[iNumSpecies];
  mp_fStmDBHCoeff = new float[iNumSpecies];
  mp_fPropDieMedDmg = new float[iNumSpecies];
  mp_fPropDieFullDmg = new float[iNumSpecies];
  mp_fPropTipUp = new float[iNumSpecies];

  //Set up our floatVal array that will extract values only for the species
  //assigned to this behavior
  p_fTempValues = new floatVal[m_iNumBehaviorSpecies];
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    p_fTempValues[i].code = mp_iWhatSpecies[i];

  //Minimum DBH for storm damage
  FillSpeciesSpecificValue( p_oElement, "st_minDBH", "st_mdVal", p_fTempValues, m_iNumBehaviorSpecies, p_oPop, true );
  //Transfer to the appropriate array buckets
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    mp_fMinStormDBH[p_fTempValues[i].code] = p_fTempValues[i].val;

  //Damage intercept for medium damage
  FillSpeciesSpecificValue( p_oElement, "st_stmDmgInterceptMed", "st_sdimVal", p_fTempValues,
       m_iNumBehaviorSpecies, p_oPop, true );
  //Transfer to the appropriate array buckets
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    mp_fStmDmgInterceptMed[p_fTempValues[i].code] = p_fTempValues[i].val;

  //Damage intercept for full damage
  FillSpeciesSpecificValue( p_oElement, "st_stmDmgInterceptFull", "st_sdifVal", p_fTempValues,
       m_iNumBehaviorSpecies, p_oPop, true );
  //Transfer to the appropriate array buckets
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    mp_fStmDmgInterceptFull[p_fTempValues[i].code] = p_fTempValues[i].val;

  //Storm intensity coefficient
  FillSpeciesSpecificValue( p_oElement, "st_stmIntensityCoeff", "st_sicVal", p_fTempValues,
       m_iNumBehaviorSpecies, p_oPop, true );
  //Transfer to the appropriate array buckets
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    mp_fStmIntensityCoeff[p_fTempValues[i].code] = p_fTempValues[i].val;

  //Storm DBH coefficient
  FillSpeciesSpecificValue( p_oElement, "st_stmDBHCoeff", "st_sdcVal", p_fTempValues, m_iNumBehaviorSpecies, p_oPop, true );
  //Transfer to the appropriate array buckets
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
    mp_fStmDBHCoeff[p_fTempValues[i].code] = p_fTempValues[i].val;

  //Proportion of trees with medium damage that die - 04/01/05 LEM
  FillSpeciesSpecificValue( p_oElement, "st_stmPropDieMedDmg", "st_spdmdVal", p_fTempValues,
       m_iNumBehaviorSpecies, p_oPop, true );
  //Transfer to the appropriate array buckets and make sure all values are
  //between 0 and 1
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
  {
    if ( p_fTempValues[i].val < 0 || p_fTempValues[i].val > 1 )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cFunction, "clStormDamageApplier::GetParameterFileData" );
      strcpy( stcErr.cMoreInfo, "All values for proportion of medium-damaged trees that die must be between 0 and 1." );
      throw( stcErr );
    }
    mp_fPropDieMedDmg[p_fTempValues[i].code] = p_fTempValues[i].val;
  }

  //Proportion of trees with full damage that die - 04/01/05 LEM
  FillSpeciesSpecificValue( p_oElement, "st_stmPropDieFullDmg", "st_spdfdVal", p_fTempValues,
       m_iNumBehaviorSpecies, p_oPop, true );
  //Transfer to the appropriate array buckets and make sure all values are
  //between 0 and 1
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
  {
    if ( p_fTempValues[i].val < 0 || p_fTempValues[i].val > 1 )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cFunction, "clStormDamageApplier::GetParameterFileData" );
      strcpy( stcErr.cMoreInfo, "All values for proportion of fully-damaged trees that die must be between 0 and 1." );
      throw( stcErr );
    }
    mp_fPropDieFullDmg[p_fTempValues[i].code] = p_fTempValues[i].val;
  }

  //Proportion of dead trees with full damage that tip-up - 04/28/05 LEM
  FillSpeciesSpecificValue( p_oElement, "st_stmPropTipUpFullDmg", "st_sptufdVal", p_fTempValues,
       m_iNumBehaviorSpecies, p_oPop, true );
  //Transfer to the appropriate array buckets and make sure all values are
  //between 0 and 1
  for ( i = 0; i < m_iNumBehaviorSpecies; i++ )
  {
    if ( p_fTempValues[i].val < 0 || p_fTempValues[i].val > 1 )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cFunction, "clStormDamageApplier::GetParameterFileData" );
      strcpy( stcErr.cMoreInfo, "All values for proportion of fully-damaged trees that tip-up must be between 0 and 1." );
      throw( stcErr );
    }
    mp_fPropTipUp[p_fTempValues[i].code] = p_fTempValues[i].val;
  }


  //Number of years damaged trees take to heal
  FillSingleValue( p_oElement, "st_numYearsToHeal", & m_iNumYearsToHeal, true );
  if ( m_iNumYearsToHeal <= 0 )
  {
    modelErr stcErr;
    stcErr.iErrorCode = BAD_DATA;
    strcpy( stcErr.cFunction, "clStormDamageApplier::GetParameterFileData" );
    strcpy( stcErr.cMoreInfo, "The number of years to heal parameter must be greater than 0." );
    throw( stcErr );
  }

  //Number of years storm-damaged snags hang on - 04/28/05 LEM
  FillSingleValue( p_oElement, "st_numYearsStormSnags", & m_iSnagYears, true );
  if ( m_iSnagYears < 0 )
  {
    modelErr stcErr;
    stcErr.iErrorCode = BAD_DATA;
    strcpy( stcErr.cFunction, "clStormDamageApplier::GetParameterFileData" );
    strcpy( stcErr.cMoreInfo, "The number of years snags last parameter cannot be less than 0." );
    throw( stcErr );
  }
}

////////////////////////////////////////////////////////////////////////////
// FormatQueryString()
// Edit history:
// -------------
// September 29, 2004 - Created (LEM)
////////////////////////////////////////////////////////////////////////////
void clStormDamageApplier::FormatQueryString()
{
  char * cQueryTemp = new char[75];
  char cQueryPiece[5]; //for assembling the search query
  int i;
  bool bSapling = false, bAdult = false, bSnag = false;

  //Do a type/species search on all the types and species
  strcpy( cQueryTemp, "species=" );
  for ( i = 0; i < m_iNumBehaviorSpecies - 1; i++ )
  {
    sprintf( cQueryPiece, "%d%s", mp_iWhatSpecies[i], "," );
    strcat( cQueryTemp, cQueryPiece );
  }
  sprintf( cQueryPiece, "%d", mp_iWhatSpecies[m_iNumBehaviorSpecies - 1] );
  strcat( cQueryTemp, cQueryPiece );

  //Find all the types
  for ( i = 0; i < m_iNumSpeciesTypeCombos; i++ )
  {
    if ( clTreePopulation::sapling == mp_whatSpeciesTypeCombos[i].iType )
    {
      bSapling = true;
    }
    else if ( clTreePopulation::adult == mp_whatSpeciesTypeCombos[i].iType )
    {
      bAdult = true;
    }
    else if ( clTreePopulation::snag == mp_whatSpeciesTypeCombos[i].iType )
    {
      bSnag = true;
    }
  }
  strcat( cQueryTemp, "::type=" );
  if ( bSapling )
  {
    sprintf( cQueryPiece, "%d%s", clTreePopulation::sapling, "," );
    strcat( cQueryTemp, cQueryPiece );
  }
  if ( bAdult )
  {
    sprintf( cQueryPiece, "%d%s", clTreePopulation::adult, "," );
    strcat( cQueryTemp, cQueryPiece );
  }
  if ( bSnag )
  {
    sprintf( cQueryPiece, "%d%s", clTreePopulation::snag, "," );
    strcat( cQueryTemp, cQueryPiece );
  }

  //Remove the last comma
  cQueryTemp[strlen( cQueryTemp ) - 1] = '\0';

  //Now put it in m_cQuery, sized correctly
  m_cQuery = new char[strlen( cQueryTemp )];
  strcpy( m_cQuery, cQueryTemp );
  delete[] cQueryTemp;
}

////////////////////////////////////////////////////////////////////////////
// GetData()
// Edit history:
// -------------
// September 29, 2004 - Created (LEM)
////////////////////////////////////////////////////////////////////////////
void clStormDamageApplier::GetData( xercesc::DOMDocument * p_oDoc )
{
  try
  {

    //Read in the parameter file data
    GetParameterFileData( p_oDoc );

    //Format our query string
    FormatQueryString();

    //Get the pointer to the storm damage grid
    mp_oStormDamageGrid = mp_oSimManager->GetGridObject( "Storm Damage" );
    if ( NULL == mp_oStormDamageGrid )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cFunction, "clStormDamageApplier::GetData" );
      strcpy( stcErr.cMoreInfo, "The storm damage applier behavior must be used with the storm behavior." );
      throw( stcErr );
    }

    m_iDmgIndexCode = mp_oStormDamageGrid->GetFloatDataCode( "dmg_index" );
    if ( -1 == m_iDmgIndexCode )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cFunction, "clStormDamageApplier::GetData" );
      strcpy( stcErr.cMoreInfo, "The grid \"Storm Damage\" is set up in an unexpected way." );
      throw( stcErr );
    }

  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clStormDamageApplier::GetData" );
    throw( stcErr );
  }
}

////////////////////////////////////////////////////////////////////////////
// Action()
// Edit history:
// -------------
// September 29, 2004 - Created (LEM)
// April 1, 2005 - Added stuff so this can kill trees (LEM)
////////////////////////////////////////////////////////////////////////////
void clStormDamageApplier::Action()
{
  try
  {
    clTreePopulation * p_oPop = ( clTreePopulation * ) mp_oSimManager->GetPopulationObject( "treepopulation" );
    clTreeSearch * p_oBehaviorTrees;
    clTree * p_oTree;
    float fStormSeverity, //storm severity in a tree's grid cell
         fX, fY, //tree's coordinates
         fDbh, //tree's dbh
         fRandom, //random number
         fMediumDmgProb, //probability of medium tree damage
         fFullDmgProb, //probability of full tree damage
         fNumYearsPerTimestep = mp_oSimManager->GetNumberOfYearsPerTimestep();
    int iTreeDamage, iYearsToHeal;
    int iSp, iTp;

    //Get the dead codes if we haven't already - have to do this here because
    //they wouldn't be available at setup time
    if ( !mp_iDeadCodes ) GetDeadCodes(p_oPop);

    //Get the trees that apply to this behavior
    p_oBehaviorTrees = p_oPop->Find( m_cQuery );

    p_oTree = p_oBehaviorTrees->NextTree();
    while ( p_oTree )
    {

      iSp = p_oTree->GetSpecies();
      iTp = p_oTree->GetType();

      //Double-check this tree's appropriateness by making sure we have a
      //non -1 return code for the stm_dmg data member
      if ( -1 == mp_iStmDmgCodes[iTp] [iSp] )
        goto nextTree;

      //If this is a snag, check to see if it needs its counter incremented or
      //if it wants removal - 04/28/05 LEM
      if ( clTreePopulation::snag == iTp )
      {

        //Get the counter
        p_oTree->GetValue( mp_iStmDmgCodes[iTp] [iSp], & iTreeDamage );

        //If the counter is 0, it's not storm-killed; so get out
        if ( 0 == iTreeDamage ) goto nextTree;

        //Get the time since damage counter and see if it's equal to the
        //snag lifespan
        iYearsToHeal = iTreeDamage > 2000 ? iTreeDamage % 2000 : iTreeDamage % 1000;
        iYearsToHeal += fNumYearsPerTimestep;
        if ( iYearsToHeal >= m_iSnagYears )
        {
          //Kill it!
          p_oPop->KillTree( p_oTree, clTreePopulation::storm );
        }
        else
        {
          //Increment the counter and put it back
          iTreeDamage += fNumYearsPerTimestep;
          p_oTree->SetValue( mp_iStmDmgCodes[iTp] [iSp], iTreeDamage );
        }

        //Nothing else to do with this tree - skip out to the next tree
        goto nextTree;
      }

      //If this tree is too small, skip out
      p_oTree->GetValue( p_oPop->GetDbhCode( iSp, iTp ), & fDbh );
      if ( fDbh < mp_fMinStormDBH[iSp] )
        goto nextTree;

      //Get the tree's existing damage
      p_oTree->GetValue( mp_iStmDmgCodes[iTp] [iSp], & iTreeDamage );

      //Get the storm severity for this tree's location
      p_oTree->GetValue( p_oPop->GetXCode( iSp, iTp ), & fX );
      p_oTree->GetValue( p_oPop->GetYCode( iSp, iTp ), & fY );
      mp_oStormDamageGrid->GetValueAtPoint( fX, fY, m_iDmgIndexCode, & fStormSeverity );

      //Was there was a storm?
      if ( fStormSeverity > 0 )
      {

        //Get the probability of medium damage
        fMediumDmgProb = exp( mp_fStmDmgInterceptMed[iSp]
             + ( mp_fStmIntensityCoeff[iSp] * fStormSeverity * pow( fDbh, mp_fStmDBHCoeff[iSp] ) ) )
             / ( 1 + exp( mp_fStmDmgInterceptMed[iSp]
             + ( mp_fStmIntensityCoeff[iSp] * fStormSeverity * pow( fDbh, mp_fStmDBHCoeff[iSp] ) ) ) );

        //Get a random number
        fRandom = clModelMath::GetRand();

        if ( fRandom <= fMediumDmgProb )
        {
          //The tree is undamaged, but may need its counter incremented
          if ( iTreeDamage > 0 )
          {
            iYearsToHeal = iTreeDamage > 2000 ? iTreeDamage % 2000 : iTreeDamage % 1000;
            iYearsToHeal += fNumYearsPerTimestep;
            if ( iYearsToHeal >= m_iNumYearsToHeal )
            {
              iTreeDamage = 0;
            }
            else
            {
              iTreeDamage += fNumYearsPerTimestep;
            }
          }
        }
        else
        {

          //The tree gets at least medium damage
          //Get the probability of full damage
          fFullDmgProb = exp( mp_fStmDmgInterceptFull[iSp]
               + ( mp_fStmIntensityCoeff[iSp] * fStormSeverity * pow( fDbh, mp_fStmDBHCoeff[iSp] ) ) )
               / ( 1 + exp( mp_fStmDmgInterceptFull[iSp]
               + ( mp_fStmIntensityCoeff[iSp] * fStormSeverity * pow( fDbh, mp_fStmDBHCoeff[iSp] ) ) ) );

          if ( fRandom <= fFullDmgProb )
          {

            //Tree gets medium damage
            //Does it die?
            if ( clModelMath::GetRand() <= mp_fPropDieMedDmg[iSp] )
            {
              //Yes it dies - if this is an adult and this behavior applies to
              //snags, create one
              if (-1 != mp_iStmDmgCodes[clTreePopulation::snag] [iSp] &&
                  clTreePopulation::adult == iTp) {
                p_oTree = p_oPop->KillTree(p_oTree, clTreePopulation::storm);
                iTreeDamage = 1000;
                p_oTree->SetValue(mp_iStmDmgCodes[clTreePopulation::snag] [iSp], iTreeDamage );
                goto nextTree;
              }
              else {
                p_oTree->SetValue( p_oPop->GetBoolDataCode( "dead", iSp, iTp ), true );
                goto nextTree;
              }
            }
            else
            {
              if ( iTreeDamage > 2000 )
              {
                //Tree already had full damage; keep that damage status
                //and reset the counter
                iTreeDamage = 2000;
              }
              else
              {
                //Tree already had no more than medium damage; so just
                //reset the counter
                iTreeDamage = 1000;
              }
            }
          }
          else
          {

            //Tree gets full damage
            //Does it die?
            if ( clModelMath::GetRand() <= mp_fPropDieFullDmg[iSp] )
            {
              //Yes it dies - if this doesn't apply to snags, or this
              //isn't an adult, just kill it
              if (-1 == mp_iStmDmgCodes[clTreePopulation::snag] [iSp] ||
                  clTreePopulation::adult != iTp) {
                p_oTree->SetValue( p_oPop->GetBoolDataCode( "dead", iSp, iTp ), true );
                goto nextTree;
              }
              else {
                p_oTree = p_oPop->KillTree(p_oTree, clTreePopulation::storm);
                //Does it tip up?
                if (clModelMath::GetRand() <= mp_fPropTipUp[iSp] ) {
                  //Yes - "kill" the snag
                  p_oTree->SetValue( p_oPop->GetBoolDataCode( "dead", iSp, clTreePopulation::snag ), true );
                }
                iTreeDamage = 2000;
                p_oTree->SetValue(mp_iStmDmgCodes[clTreePopulation::snag] [iSp], iTreeDamage );
                goto nextTree;
              }
            }
            else
            {
              iTreeDamage = 2000;
            }
          }
        }
      }
      else
      {

        //There was no storm, but the tree may need decrementing
        if ( iTreeDamage > 0 )
        {
          iYearsToHeal = iTreeDamage > 2000 ? iTreeDamage % 2000 : iTreeDamage % 1000;

          iYearsToHeal += fNumYearsPerTimestep;
          if ( iYearsToHeal >= m_iNumYearsToHeal )
          {
            iTreeDamage = 0;
          }
          else
          {
            iTreeDamage += fNumYearsPerTimestep;
          }
        } else goto nextTree;
      }

      p_oTree->SetValue( mp_iStmDmgCodes[iTp] [iSp], iTreeDamage );

    nextTree:
      p_oTree = p_oBehaviorTrees->NextTree();
    }
  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clStormDamageApplier::Action" );
    throw( stcErr );
  }
}

